/*-----------------------------------------------------------------------*
  Project:  Nintendo Augmented Reality Library.
  File:     narDistortion.h

  Copyright (C)2011-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo and/or its licensed developers
  and are protected by national and international copyright laws. They
  may not be disclosed to third parties or copied or duplicated in any
  form, in whole or in part, without the prior written consent of
  Nintendo.
  The content herein is highly confidential and should be handled
  accordingly.
*-----------------------------------------------------------------------*/

/// @cond DEV
/**
 *  @file
 *  @author Satoru Osako
 */

#ifndef NAR_DISTORTION_H__
#define NAR_DISTORTION_H__

#include "narPort.h"
#include "narVec2.h"

namespace mw { namespace nar
{
    /** Yɂcݕ␳NX
     */
    class Distortion_cl
    {
    public:
        /** RXgN^
         *  @param x        摜̒SxW
         *  @param y        摜̒SyW
         *  @param factor   cȎ␳p̌W
         *
         *  @par factor̋ߕ
         *  摜Ŝ̍\f$ Y \f$ɑ΂摜[̃Y\f$ \Delta y \f$^ꂽƂA
         *  摜̏c䂪3:4ƂāAʂ̉E[̓_ṔA\f$ P( 2Y/3,Y/2 ) \f$ŗ^B
         *  ܂A̒̋̓́A
         *  \f[
         *      d^2 = (\frac{2Y}{3})^2 + ( \frac{Y}{2} )^2
         *  \f]
         *  łBāA_PɂĈړ_P'YẂA
         *  \f{eqnarray*}{
         *      \frac{Y}{2} - \Delta y &=& ( 1 - \frac{1}{factor}( ( \frac{2Y}{3} )^2 + ( \frac{Y}{2} )^2 ) )  \frac{Y}{2} \\
         *      \frac{Y}{2} - \Delta y &=& \frac{Y}{2} - \frac{Y}{2factor}((\frac{2Y}{3})^2+(\frac{Y}{2})^2) \\
         *      \frac{Y}{2factor}((\frac{2Y}{3})^2+(\frac{Y}{2})^2) &=& \Delta y \\
         *      \frac{Y}{2}((\frac{2Y}{3})^2+(\frac{Y}{2})^2) &=& factor \Delta y \\
         *      factor &=& \frac{Y}{2\Delta y}((\frac{2Y}{3})^2+(\frac{Y}{2})^2)
         *  \f}
         *  \f$ \Delta y = \frac{a}{100}Y \f$ ŗ^ƂAfactor̒l
         *  \f{eqnarray*}{
         *      factor &=& \frac{Y}{2\frac{a}{100}Y}((\frac{2Y}{3})^2+(\frac{Y}{2})^2) \\
         *             &=& \frac{50}{a}(\frac{4Y^2}{9}+\frac{Y^2}{4}) \\
         *             &=& \frac{50}{a}\cdot\frac{16Y^2 + 9Y^2}{36} \\
         *             &=& \frac{25\cdot 25Y^2}{18a} \\
         *             &=& \frac{625Y^2}{18a} \\
         *  \f}
         *  œB
         */
        Distortion_cl( f32 x, f32 y, f32 factor )
             : m_Center( x, y ), m_Factor( factor )
        {}

        /** cȎ̎l痝zl߂
         */
        void GetIdealFromObserv( Vec2F_st & r_Ideal, const Vec2F_st & cr_Observ ) const;
        /** zlcȎ̒l߂
         */
        void GetObservFromIdeal( Vec2F_st & r_Observ, const Vec2F_st & cr_Ideal ) const;
    private:
        Vec2F_st    m_Center;   ///< 摜̒
        f32         m_Factor;   ///< W
    };
}
}
#endif
/// @endcond
