/*-----------------------------------------------------------------------*
  Project:  Nintendo Augmented Reality Library.
  File:     narEdgeLine.h

  Copyright (C)2011-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo and/or its licensed developers
  and are protected by national and international copyright laws. They
  may not be disclosed to third parties or copied or duplicated in any
  form, in whole or in part, without the prior written consent of
  Nintendo.
  The content herein is highly confidential and should be handled
  accordingly.
*-----------------------------------------------------------------------*/

/// @cond DEV
/**
 *  @file
 *  @author Satoru Osako
 */
#ifndef NAR_EDGE_LINE_H__
#define NAR_EDGE_LINE_H__

#include "narVec2.h"

namespace mw { namespace nar
{
    /**
     *  [̒NX
     */
    class EdgeLine_cl
    {
    public:
        EdgeLine_cl()
        {
            Init();
        }

        void Init();

        /** _̎Zo
         */
        bool GetIntersection( Vec2F_st * p_Point, const EdgeLine_cl & r_Edge ) const;

        /** ̉
         */
        void Extend( const EdgeLine_cl & r );
        /** ̉(Jn_)
         */
        void ExtendStart( const EdgeLine_cl & r );
        /** ̉(I_)
         */
        void ExtendEnd( const EdgeLine_cl & r );
        
        f32   GetA() const { return m_A; }
        f32   GetB() const { return m_B; }
        f32   GetC() const { return m_C; }

        /** Jn_o^
         */
        void SetStart( s32 x, s32 y )
        {
            m_Start.Set( x, y );
        }

        /** Jn_o^
         */
        void SetStart( const Vec2I_st & cr )
        {
            SetStart( cr.X(), cr.Y() );
        }

        /** I_o^
         */
        void SetEnd( s32 x, s32 y )
        {
            m_End.Set( x, y );
        }

        /** I_o^
         */
        void SetEnd( const Vec2I_st & cr )
        {
            SetEnd( cr.X(), cr.Y() );
        }

        /** ̓擾
         */
        s32  GetSquareLength() const
        {
            return GetrStart().GetSquareDistance( GetrEnd() );
        }

        /** Tv_ǉ
         */
        void AddSample( s32 x, s32 y );
        /** Tv_ǉ
         */
        void AddSample( const Vec2I_st & cr )
        {
            AddSample( cr.X(), cr.Y() );
        }

        /** startend璼vZ
         */
        bool CalcEquation();
        /**
         *  ŏ@ɂ蒼𐄒
         */
        bool SolveLinearLeastSquares();
        /**
         *  _ƒ̋߂B
         */
        f32 GetDistance( const Vec2F_st &r ) const;
        /**
         *  _ƒ̋߂B
         */
        f32 GetDistance( const Vec2I_st &r ) const;

        /** GbW(̂)
         */
        bool CheckSameEdgeDir( const EdgeLine_cl & cr, f32 * p_Sin, s32 * p_Dot ) const;

        /** GbW
         */
        bool CheckSameEdge( const EdgeLine_cl & cr, f32 * p_Sin, s32 * p_Dot ) const;

    private:
        ///   ax + by + c = 0  a
        f32         m_A;
        ///   ax + by + c = 0  b
        f32         m_B;
        ///   ax + by + c = 0  c
        f32         m_C;

        Vec2I_st    m_Start;
        Vec2I_st    m_End;
        s32         m_SampleNum;
        s32         m_SumSquareXi;
        s32         m_SumSquareYi;
        s32         m_SumXi;
        s32         m_SumYi;
        s32         m_SumXiYi;

        const Vec2I_st & GetrStart() const { return m_Start; }
        const Vec2I_st & GetrEnd()  const { return m_End; }
        void addSums( const EdgeLine_cl & cr );
    };
}
}
#endif
/// @endcond
