/*-----------------------------------------------------------------------*
  Project:  Nintendo Augmented Reality Library.
  File:     narImage.h

  Copyright (C)2011-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo and/or its licensed developers
  and are protected by national and international copyright laws. They
  may not be disclosed to third parties or copied or duplicated in any
  form, in whole or in part, without the prior written consent of
  Nintendo.
  The content herein is highly confidential and should be handled
  accordingly.
*-----------------------------------------------------------------------*/

/**
 *  @file
 *  @author Satoru Osako
 */

#ifndef NAR_IMAGE_H__
#define NAR_IMAGE_H__

#include "narPort.h"
#include "narMarkerTemplate.h"
#include "narMesh.h"

namespace mw { namespace nar
{
    struct Color_st;

    /**	J[h̗֊s@܂B
     */
    enum Border_en
    {
        e_BorderWhiteBlack,	///<	ɍñfUC
        e_BorderBlackWhite,	///<	ɔñfUC
        e_BorderEnd
    };
    
    /** }[J[oΏۂƂȂ摜NXłB
     */
    class Image_cl
    {
    public:
        /// ֊soX̃Xebvl
        static const u8 msc_DefaultXStep = 8;
        /// ֊soỸXebvl
        static const u8 msc_DefaultYStep = 16;
        
        /**
         *  @param w            摜̕
         *  @param h            摜̍
         *  @param p_Work       [N̈ւ̃|C^ ([N̑傫 NAR_IMAGE_WORK_SIZE_4TH }NŎ擾Ă)
         *  @param p_DebugImg   fobO\摜̈ւ̃|C^ (NAR_DEBUG__==1̎)
         */
        Image_cl( u16 w, u16 h, u32 * p_Work, u16 * p_DebugImg = 0 )
             : m_Width( w ), m_Height( h ), m_XStep( msc_DefaultXStep ), m_YStep( msc_DefaultYStep ), m_Border( 0 ), mcp_Image( 0 ), mp_CheckFlags( p_Work )
#if NAR_DEBUG__
               , mp_DebugImage( p_DebugImg )
#endif
        {
            NAR_NULL_TASSERT( p_Work );
#if ! NAR_DEBUG__
            ( ( void )&p_DebugImg );
#endif
        }

        /**
         *  摜o^܂
         *  @param cp_Image     o^摜ւ̃|C^
         */
        void SetImage( const u16 * cp_Image )
        {
            NAR_NULL_TASSERT( cp_Image );
            mcp_Image = cp_Image;
        }

        /**
         *  摜̍擾܂
         */
        u16 GetHeight() const { return m_Height; }

        /**
         *  摜̕擾܂
         */
        u16 GetWidth()  const { return m_Width; }

        /**
         *  T߂x̃Xebv擾܂B
         *  o}[J[̉̉摜̕Ɋ֌W܂B
         */
        u8  GetXStep()  const { return m_XStep; }

        /**
         *  T߂ỹXebv擾܂B
         *  o}[J[̉摜Y̑傫Ɋ֌W܂B
         */
        u8  GetYStep()  const { return m_YStep; }

        /**
         *  }[J[̉Tۂx̃Xebvݒ肵܂B
         *
         *  ̒l傫ƁAx͌サ܂A摜ŁAʂĂꍇɁA
         *  }[J[̉ołȂ\oĂ܂B
         *  ftHg̒ĺA nar::Image_cl::msc_DefaultXStep Œ`Ă܂B
         *  @param x Xebv
         */
        void    SetXStep( u8 x )
        {
            NAR_MIN_TASSERT( x, 1 );
            m_XStep = x;
        }

        /**
         *  T߂ỹXebvݒ肵܂B
         *
         *  ̒l傫ƁAx͌サ܂A摜ŁAʂĂꍇɁA
         *  }[J[̉ołȂ\oĂ܂B
         *  ftHg̒ĺA nar::Image_cl::msc_DefaultYStep Œ`Ă܂B
         *  @param y Xebv
         */
        void    SetYStep( u8 y )
        {
            NAR_MIN_TASSERT( y, 1 );
            m_YStep = y;
        }
        
        /** QƓ_̐F擾܂
         *  @param x XW
         *  @param y YW
         *	@retval	RGB5551`̐FԂ܂B
         */
        u16     GetColor( u32 x, u32 y ) const
        {
#if NAR_PIXEL_FMT == NAR_PIXEL_FMT_RGB555
            return static_cast< u16 >( mcp_Image[ y * m_Width + x ] );
#endif
#if NAR_PIXEL_FMT == NAR_PIXEL_FMT_YUV422_BATCH
            int pos = y * m_Width + x;
            u8 Y = static_cast< u8 >( mcp_Image[ pos ] & 0xff );
            u8 U = static_cast< u8 >( mcp_Image[ pos & ~1 ] >> 8 & 0xff );
            u8 V = static_cast< u8 >( mcp_Image[ ( pos & ~1 ) + 1 ] >> 8 & 0xff );
            return ConvYUVtoRGB5551( Y, U, V );
#endif
        }

        /**	֊s̔@ݒ肵܂B
         */
        void		SetBorder( Border_en b )
        {
            m_Border	= b;
        }

        /**	֊s̔@擾܂B
         */
        Border_en	GetBorder()
        {
            return static_cast< Border_en >( m_Border );
        }

        /** Cӂ̂S_iqɕĊe_̃J[擾܂B
         *  @tparam N               ӂ̕(MeshDiv_e)
         *  @param cap_Points       Cӂ̂S_
         *  @param r_MeshColorData  J[̊i[
         */
        template< MeshDiv_e N >
        void     PickupMeshData( const Vec2F_st * cap_Points, MeshColorData_tc< N > & r_MeshColorData ) const
        {
            // S_烁bVvZ
            MeshStrictWork_tc< N > work;
            MeshStrict_cl mesh( work );
            mesh.precalc( cap_Points );
#if NAR_DEBUG__
            //  fobOp\
            mesh.drawInfo( *this, cap_Points, false );
#endif
            mesh.pickupFast( *this, r_MeshColorData );
        }

        /** Cӂ̂S_iqɕĊe_̃J[ƍWl擾܂B
         *  @tparam N                ӂ̕iMeshDiv_ej
         *  @param cap_Points       Cӂ̂S_
         *  @param r_MeshColorData  J[̊i[
         *  @param r_MeshVtxData    W̊i[
         */
        template< MeshDiv_e N >
        void     PickupMeshData( const Vec2F_st * cap_Points, MeshColorData_tc< N > & r_MeshColorData, MeshVtxData_tc< N > & r_MeshVtxData ) const
        {
            // S_烁bVvZ
            MeshStrictWork_tc< N > work;
            MeshStrict_cl mesh( work );
            mesh.precalc( cap_Points );
#if NAR_DEBUG__
            //  fobOp\
            mesh.drawInfo( *this, cap_Points, false );
#endif
            mesh.pickupFast( *this, r_MeshColorData, r_MeshVtxData );
        }

        /// @cond DEV
        /** QƓ_͂̕ϐF擾
         *  @param x XW
         *  @param y YW
         */
        void     GetAvgColor( u32 x, u32 y, Color_st * p_Color ) const;

        /** ̂P_݂̂ŕϐFƂĐF擾
         *  @param x XW
         *  @param y YW
         */
        void     GetAvgColorFast( u32 x, u32 y, Color_st * p_Color ) const;

        /** Ԑ擾(0,31)
         *  @param color F(RGB5A1)
         */
        static u8 GetR( u16 color )
        {
            return static_cast< u8 >( color >> NAR_RGB5A1_R_SHIFT & 0x1f );
        }

        /** ΐ擾(0,31)
         *  @param color F(RGB5A1)
         */
        static u8 GetG( u16 color )
        {
            return static_cast< u8 >( color >> NAR_RGB5A1_G_SHIFT & 0x1f );
        }

        /** 擾(0,31)
         *  @param color F(RGB5A1)
         */
        static u8 GetB( u16 color )
        {
            return static_cast< u8 >( color >> NAR_RGB5A1_B_SHIFT & 0x1f );
        }

        /** PPxvZ(0,124)
         */
        static u8 GetSimpleYRaw( u8 r, u8 g, u8 b )
        {
            return static_cast< u8 >( r + ( g << 1 ) + b );
        }

        /** PPxvZ(0,255)
         */
        u8 GetSimpleYRaw( u32 x, u32 y ) const
        {
#if NAR_PIXEL_FMT == NAR_PIXEL_FMT_RGB555
            u16 color = GetColor( x, y );
            return GetSimpleYRaw( GetR( color ), GetG( color ), GetB( color ) );
#endif
#if NAR_PIXEL_FMT == NAR_PIXEL_FMT_YUV422_BATCH
            u8 ret = static_cast< u8 >( mcp_Image[ y * m_Width + x ] & 0xff );
            if ( m_Border )
                return ~ret;
            else
                return ret;
#endif
        }

        /** sNZɃ`FbNtO𗧂Ă
         */
        void Checked( u32 x, u32 y )
        {
            register u32 pixelNo    = y * m_Width + x;
            register u32 flagNo     = pixelNo >> 5;
            register u32 bitNo      = pixelNo & 0x1f;
            mp_CheckFlags[ flagNo ] |= ( 1 << bitNo );
        }

        /** sNZ`FbNꂽǂԂ
         */
        bool IsChecked( u32 x, u32 y ) const
        {
            register u32 pixelNo    = y * m_Width + x;
            register u32 flagNo     = pixelNo >> 5;
            register u32 bitNo      = pixelNo & 0x1f;
            return 0 != ( mp_CheckFlags[ flagNo ] & ( 1 << bitNo ) );
        }

        /** `FbNtONA
         */
        void ClearChecked()
        {
            ClearFast( mp_CheckFlags, static_cast< u32 >( (( m_Width * m_Height + 31 ) & ~31) >> 3 ) );
        }

        /** YUVf[^RGB555ɕϊ
         */
        static u16  ConvYUVtoRGB5551( u8 y, u8 u, u8 v );


#if NAR_DEBUG__
#if NAR_CTR_NATIVE_DEBUG_IMAGE
        /** DmplCeButH[}bgɂeNZʒu̎擾
         *  ̃NXň摜́AEɁAxWAyWƂ̂łɍ킹B
         */
        u32     GetDmpNativeTexelPos( u32 x, u32 y ) const;
#endif


        /** fobO摜ɐFݒ肷
         *  @param x XW
         *  @param y YW
         */
        void    SetColorDebug( u32 x, u32 y, u16 color ) const
        {
            if ( mp_DebugImage && x < m_Width && y < m_Height )
            {
#if NAR_CTR_NATIVE_DEBUG_IMAGE
                int idx = GetDmpNativeTexelPos( x, y );
#else
                int idx = y * m_Width + x;
#endif
                if ( 0 == mp_DebugImage[ idx ] )
                    mp_DebugImage[ idx ] = color;
            }
        }

        /** fobO摜NA
         */
        void    ClearDebugImage()
        {
            if ( mp_DebugImage )
                ClearFast( mp_DebugImage, m_Width * m_Height * sizeof( u16 ) );
        }

        /** `悷
         */
        void    DrawLineSegmentOnDebug( const Vec2F_st & cr_Start, const Vec2F_st & cr_End, u16 color ) const;

        /** `悷B
         */
        void    DrawLineOnDebug( f32 a, f32 b, f32 c, u16 color ) const;

        /** `悷B
         */
        void    DrawArrowOnDebug( const Vec2I_st & cr_Start, const Vec2I_st & cr_End, u16 color ) const;

        /** \`悷
         */
        void    DrawPlusMarkOnDebug( const Vec2I_st & r_Point, u16 color ) const;

#endif
        /// @endcond
        /** fobO摜̈ݒ肷B(NAR_DEBUG__==1̎̂ݗL)
         */
        void    SetDebugImage( u16 * p_Image )
        {
#if NAR_DEBUG__
            mp_DebugImage   = p_Image;
#else
            ((void)&p_Image);
#endif
        }
    private:
        /// 摜f[^̕
        u16     m_Width;
        /// 摜f[^̍
        u16     m_Height;
        /// x̃Xebv
        /// o}[J[̉̉摜̕Ɋ֌W܂B
        u8      m_XStep;
        /// ỹXebv
        /// o}[J[̉摜Y̑傫Ɋ֌W܂B
        u8      m_YStep;

        ///	֊s
        u8		m_Border;

        u8     	pad;
        
        /// 摜f[^ւ̃|C^
        const u16 * mcp_Image;
        /// sNZ`FbNtO
        u32 *   mp_CheckFlags;
#if NAR_DEBUG__
        u16 *   mp_DebugImage;
#endif
    };

    /**
     *  [NTCYvZp}N(32bitf[^)
     */
#define NAR_IMAGE_WORK_SIZE_4TH( w, h ) (((w)*(h)+31)>>5)
}
}

#endif
