/*-----------------------------------------------------------------------*
  Project:  Nintendo Augmented Reality Library.
  File:     narMarker.h

  Copyright (C)2011-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo and/or its licensed developers
  and are protected by national and international copyright laws. They
  may not be disclosed to third parties or copied or duplicated in any
  form, in whole or in part, without the prior written consent of
  Nintendo.
  The content herein is highly confidential and should be handled
  accordingly.
*-----------------------------------------------------------------------*/
/**
 *  @file
 *  @author Satoru Osako
 */

#ifndef NAR_MARKER_H__
#define NAR_MARKER_H__

#include "narConfig.h"
#include "narPort.h"
#include "narVec2.h"
#include "narVec3.h"
#include "narProjection.h"
#include "narTransformation.h"
#include "narMarkerTemplate.h"
#include "narDistortion.h"

namespace mw { namespace nar
{
    class Distortion_cl;
    /** o}[J[NXłB
     */
    class Marker_cl
    {
    public:
        Marker_cl() : mcp_PatternVtxData( 0 )
        {}

        /** ID擾
         *  @retval }[J[ID(G[-1)
         */
        s32 GetID() const { return mcp_Template ? mcp_Template->id : -1; }

        /** XRA擾܂B
         *  @retval XRAB͈͂́A-1.0 <= score <= 1.0
         */
        f32 GetScore() const { return m_Score; }

        /** }[J[̌擾܂B
         *  @retval e_MarkerDir_Top, e_MarkerDir_Left, e_MarkerDir_Bottom, e_MarkerDir_Right̉ꂩ
         */
        s32 GetDir() const { return m_Dir; }

        /** ϊs擾
         */
        Transformation_cl & GetrTransformation() { return m_Trans; }

        /** Wς݂ǂ
         */
        bool IsEstimated() const { return m_Estimated; }

        /** }[J[̍ӂ̃[hWnł̒
         */
        f32 GetSideLength() const
        {
            return mcp_Template ? mcp_Template->sideLength : 0 ;
        }

        /** }[J[̉/c̃AXyNg
         */
        f32 GetAspectRatio() const
        {
            return mcp_Template ? mcp_Template->aspectRatio : 0 ;
        }

        /** }[J[ʂ̖@xNg
         */
        const Vec3F_st & GetrNormal() { return m_Trans.GetrYAxis(); } 

        /** Öʒuk}
         *  @param  sim ȑÕ}[J[ʒůe_Ƃ̘̋̓a̒lȉȂ瓯Ƃ݂Ȃ
         *  @param  far ̒lȉȂԂ
         */
        void SuppressShake( f32 sim, f32 far );

        /** }[J[pp^[̊iqWf[^̃|C^擾܂B
         *
         *  - mw::nar::MarkerAnalizableData_tcgă}[J[oꍇ̂݃f[^擾ł܂
         *  - mw::nar::MarkerData_tcŌoꍇ͗płȂ̂łӂ
         *  @retval }[J[pp^[̊iqWf[^̃|C^
         */
        const PatternVtxData_cl * GetpPatternVtxData()
        {
            return mcp_PatternVtxData;
        }

        /// @cond DEV
        /** 
         */
        void Init( const MarkerTemplate_st * cp_Temp,
                   const Vec2F_st * cap_Points,
                   const PatternVtxData_cl * cp_PatternVtxData,
                   s32 dir,
                   f32 c_Score );


        ///
        void SetEstimated( bool b ) { m_Estimated = b; }

        const Vec2F_st * GetpDetectedVtx() const
        {
            return ma_VtxDetected;
        }

        void SetDetectedVtx( s32 idx, f32 x, f32 y )
        {
            ma_VtxDetected[ idx ].Set( x, y );
        }

        const Vec3F_st & GetrVtxOnRefPlane( s32 idx ) const
        {
            return ma_VtxOnRefPlane[ idx ];
        }
        
        Marker_cl * GetpPrev() { return mp_Prev; }
        void SetPrev( Marker_cl * p_Prev ) { mp_Prev = p_Prev; }
        Marker_cl * GetpNext() { return mp_Next; }
        void SetNext( Marker_cl * p_Next ) { mp_Next = p_Next; }

        /** ^ꂽ}[J[Ƃ̑ԂB
         */
        f32 GetSumOfEachVtxSquareDistance( const Marker_cl & cr ) const;

        /** ^ꂽ}[J[ƒ_𕽋ωB
         */
        void AverageVtx( const Marker_cl & cr );

        /** Qƕʏ̓_߂
         */
        void    CalcVtxOnRefPlane( const Projection_cl & cr_Proj )
        {
            for ( int vtx = 0; vtx < 4; vtx++ )
            {
                cr_Proj.GetVtxOnReferencePlane( ma_VtxOnRefPlane[ vtx ], ma_VtxDetected[ vtx ] );
            }
        }

        /** Qƕʏ̓_߂
         */
        void    CalcVtxOnRefPlane( const Projection_cl & cr_Proj, const Distortion_cl & cr_Dist )
        {
            for ( int vtx = 0; vtx < 4; vtx++ )
            {
                Vec2F_st ideal;
                cr_Dist.GetIdealFromObserv( ideal, ma_VtxDetected[ vtx ] );
                cr_Proj.GetVtxOnReferencePlane( ma_VtxOnRefPlane[ vtx ], ideal );
            }
        }

        /** 肳ꂽϊs]
         */
        f32 EvaluateMatrix() const;

        /** oꂽ_f[^ɋ߂Â悤ɁA]𒲐
         *  @param targetErr    ڕWƂG[lie_Ԃ̘̋̓aj
         *  @param maxCnt       ő
         *
         *  Wn̊eɑ΂ď]āAG[lȂĂ܂B
         */
        f32 AdjustRotate( f32 targetErr, s32 maxCnt );

        /**
         */
        Vec3F_st & GetrYAxisMove() { return m_YAxisMove; }

        ///
        void SetYAxisMoveRec( bool b ) { m_YAxisMoveRec = b; }

        /**
         *  @pre mp_Prev && m_YAxisMoveRec == true
         */
        bool IsSimilarYAxisRotate();

        f32     GetInterpolationCoefficient() const { return m_InterpolationCoefficient; }
        void    SetInterpolationCoefficient( f32 t ) { m_InterpolationCoefficient = t; }

        /// 4{_̋̓
        f32 GetSqareDistanceQuadMid( const Vec2F_st * cap_Points );
        /// 4{_̋̓
        f32 GetSqareDistanceQuadMid( const Marker_cl & cr )
        {
            return GetSqareDistanceQuadMid( cr.ma_VtxDetected );
        }
        
        /// ʐς2{
        static f32 GetDoubleArea( const Vec2F_st * cap_Points );
        /// ʐς2{
        f32 GetDoubleArea()
        {
            return GetDoubleArea( ma_VtxDetected );
        }
        
        /// ֊sdȂĂ邩ǂԂ
        bool IsOverlapped( const Vec2F_st * cap_Points )
        {
            return GetDoubleArea( cap_Points ) * 2.f > GetSqareDistanceQuadMid( cap_Points );
        }

        /// ԒtOZbg
        void SetInterpolated( bool b ) { m_Interpolated = b; }
        
        /// @endcond
    private:
        /// oꂽ_
        /// AAEAȄŊi[Ă
        Vec2F_st                    ma_VtxDetected[ 4 ];
        /// QƖʏ̒_
        Vec3F_st                    ma_VtxOnRefPlane[ 4 ];
        /// õXRA
        f32                         m_Score;
        /// o
        s32                         m_Dir;
        /// ev[gւ̃|C^
        const MarkerTemplate_st *   mcp_Template;
        /// ϊs
        Transformation_cl           m_Trans;

        /// Ot[̉摜̓ID̃}[J[ւ̃N
        Marker_cl *         mp_Prev;
        Marker_cl *         mp_Next;

        /// ps𐄒肵ǂB
        bool                m_Estimated;

        /// Õ}[J[W(0,1,0)̓ۑĂ邩ǂ
        bool                m_YAxisMoveRec;

        bool                m_Interpolated;

        u8                  pad;

        /// ԌW
        f32                 m_InterpolationCoefficient;

        /// Õ}[J[W(0,1,0)獡̃}[J[W(0,1,0)ւ̃xNg
        Vec3F_st            m_YAxisMove;

        /// sbNAbv|Cg擾p̈ւ̃|C^
        const PatternVtxData_cl * mcp_PatternVtxData;
    };
}
}

#endif
