/*-----------------------------------------------------------------------*
  Project:  Nintendo Augmented Reality Library.
  File:     narMarkerDetector.h

  Copyright (C)2011-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo and/or its licensed developers
  and are protected by national and international copyright laws. They
  may not be disclosed to third parties or copied or duplicated in any
  form, in whole or in part, without the prior written consent of
  Nintendo.
  The content herein is highly confidential and should be handled
  accordingly.
*-----------------------------------------------------------------------*/
/**
 *  @file
 *  @author Satoru Osako
 */

#ifndef NAR_MARKER_DETECTOR_H__
#define NAR_MARKER_DETECTOR_H__

#include "narConfig.h"
#include "narPort.h"
#include "narVec2.h"
#include "narEdgeLine.h"
#include "narMarker.h"
#include "narFlags.h"

namespace mw { namespace nar
{
    class Image_cl;

    enum
    {
        e_MarkerEdgeNum = 30    ///< ołGbW̐i}[J[j
    };

    /** o}[J[̃XgۃNXłB
     */
    class MarkerList_ac
    {
    public:
        MarkerList_ac()
             : m_Num( 0 ), m_MultiId( false )
        {}
        /** idxԖڂ̃}[J[Ԃ܂
         *  @retval 0Ȃidxs
         */
        virtual Marker_cl * GetpMarker( u32 idx ) = 0;

        /** w肳ꂽID̃}[J[Ԃ܂
         *  @param id ID
         */
        virtual s32 GetMarkerNumWithID( s32 id ) = 0;

        /** Xg̃}[J[̌Ԃ܂
         */
        u16     GetMarkerNum() { return m_Num; }

        /// @cond DEV

        /** tǂԂ܂
         */
        virtual bool IsFull() = 0;

        virtual Marker_cl * GetpNewMarker() = 0;

        virtual PatternVtxData_cl * GetpWritableVtxData() = 0;

        void Initialize() { m_Num = 0; m_MultiId = false; }

        void IncMarkerNum() { m_Num++ ; }

        bool IsMultiId() { return m_MultiId; }
        void SetMultiId( bool b ) { m_MultiId = b; }
        /// @endcond
    private:
        u16             m_Num;
        u16             m_MultiId;
    };
    /** o}[J[̃XgNXłB
     *  @tparam N    oł}[J[̌w肵܂B
     */
    template < s32 N >
    class MarkerList_tc : public MarkerList_ac
    {
    public:
        MarkerList_tc()
        {}

        virtual Marker_cl * GetpMarker( u32 idx )
        {
            if ( idx >= N ) return 0;
            else            return &ma_Marker[ idx ];
        }

        virtual s32 GetMarkerNumWithID( s32 id )
        {
            int count = 0;
            for ( int idx = GetMarkerNum() - 1; idx >= 0; idx-- )
            {
                if ( ma_Marker[ idx ].GetID() == id )
                    count++;
            }
            return count;
        }

        /// @cond DEV

        virtual bool        IsFull() { return GetMarkerNum() == N; }

        virtual Marker_cl * GetpNewMarker()
        {
            int num = GetMarkerNum();
            if ( num >= N ) return 0;
            else
            {
                IncMarkerNum();
                return &ma_Marker[ num ];
            }
        }

        virtual PatternVtxData_cl * GetpWritableVtxData(){ return 0; }

        /// @endcond
    private:
        Marker_cl           ma_Marker[ N ];
    };

    /** o}[J[̃XgNXłB
     *  @tparam N    oł}[J[̌w肵܂B
     */
    template < s32 N >
    class MarkerAnalizableList_tc : public MarkerList_ac
    {
    public:
        MarkerAnalizableList_tc()
        {}

        virtual Marker_cl * GetpMarker( u32 idx )
        {
            if ( idx >= N ) return 0;
            else            return &ma_Marker[ idx ];
        }

        virtual s32 GetMarkerNumWithID( s32 id )
        {
            int count = 0;
            for ( int idx = GetMarkerNum() - 1; idx >= 0; idx-- )
            {
                if ( ma_Marker[ idx ].GetID() == id )
                    count++;
            }
            return count;
        }

        /// @cond DEV

        virtual bool        IsFull() { return GetMarkerNum() == N; }

        virtual Marker_cl * GetpNewMarker()
        {
            int num = GetMarkerNum();
            if ( num >= N ) return 0;
            else
            {
                IncMarkerNum();
                return &ma_Marker[ num ];
            }
        }

        virtual PatternVtxData_cl * GetpWritableVtxData()
        {
            // ̍Wۑ
            int num = GetMarkerNum();
            if ( num >= N ) return 0;
            else
            {
                return &ma_PatternVtxData[ num ];
            }
        }

        /// @endcond
    private:
        Marker_cl           ma_Marker[ N ];
        PatternVtxData_cl   ma_PatternVtxData[ N ];
    };

    /** o}[J[ێ钊ۃNXłB
     */
    class MarkerData_ac
    {
    public:
        MarkerData_ac()
             : m_Detecting( 0 )
        {}

        /// o}[J[̃Xg擾
        virtual MarkerList_ac & GetrDetectingMarkerList() = 0;
        /// O񌟏o}[J[̃Xg擾
        virtual MarkerList_ac & GetrLastMarkerList() = 0;

        /// @cond DEV

        /// o}[J[ǉ
        virtual bool AddMarker( const MarkerTemplate_st * cp_Temp,
                                const Vec2F_st * cap_Points,
                                s32 dir,
                                const f32 c_Score,
                                const PatternVtxData_cl * cp_VtxData ) = 0;

        /// Xgւ
        void SwapList() { m_Detecting ^= 1; }
    protected:
        u8         getDetecting() { return m_Detecting; }
        /// @endcond
    private:
        /// õXgoĂ
        u8                  m_Detecting;
        u8                  apad[3];
    };

    /** o}[J[ێ邽߂̃NXłB
     *  @tparam T    ێ郊Xǧ^w肵܂B
     *  @tparam N    oł}[J[̌w肵܂B
     */
    template < typename T, s32 N >
    class MarkerContainer_tc : public MarkerData_ac
    {
    public:
        virtual MarkerList_ac & GetrDetectingMarkerList()
        {
            return ma_MarkerLists[ getDetecting() ];
        }
        virtual MarkerList_ac & GetrLastMarkerList()
        {
            return ma_MarkerLists[ getDetecting()^1 ];
        }

        /// @cond DEV
        
        /// o}[J[ǉ
        virtual bool AddMarker( const MarkerTemplate_st * cp_Temp,
                                const Vec2F_st * cap_Points,
                                s32 dir,
                                const f32 c_Score,
                                const mw::nar::PatternVtxData_cl * cp_VtxData )
        {
            MarkerList_ac & r_New  = GetrDetectingMarkerList();

            //  ̈o^悤ƂĂȂ`FbN
            for ( int idx = r_New.GetMarkerNum() - 1; idx >= 0; idx-- )
            {
                Marker_cl * p_Marker = r_New.GetpMarker( idx );
                if ( p_Marker->GetID() == cp_Temp->id )
                {
                    if ( p_Marker->IsOverlapped( cap_Points ) )
                    {
                        //  dȂĂXRA`FbNčΏ㏑
                        if ( p_Marker->GetScore() < c_Score )
                        {
                            p_Marker->Init( cp_Temp, cap_Points, cp_VtxData, dir, c_Score );
                            return true;
                        }
                        return false;
                    }
                    else
                    {
                        //  }[J[tO
                        r_New.SetMultiId( true );
                    }
                }
            }

            //  VKo^
            Marker_cl * p_New = r_New.GetpNewMarker();

            if ( p_New )
            {
                //  V}[J[o^łȂ
                p_New->Init( cp_Temp, cap_Points, cp_VtxData, dir, c_Score );
                return true;
            }
            return false;
        }
        /// @endcond
    private:
        /** }[J[Xg
         *
         *  Ǒoʂ␳ɗp邽߁AۑĂ
         */
        T  ma_MarkerLists[ 2 ];
    };

    /** o}[J[ێNXłB
     *  @tparam N    oł}[J[̌w肵܂B
     */
    template< s32 N >
    class MarkerData_tc : public MarkerContainer_tc< mw::nar::MarkerList_tc< N >, N >
    {};

    /** o}[J[̍W͉\ȏԂŕێNXłB
     *  @tparam N    oł}[J[̌w肵܂B
     */
    template< s32 N >
    class MarkerAnalizableData_tc : public MarkerContainer_tc< mw::nar::MarkerAnalizableList_tc< N >, N >
    {};

    /** }[J[oۂɎgpA[N̈ۃNXłB
     */
    class MarkerDetectWork_ac
    {
    public:
        /// @cond DEV
        virtual bool Extract( const Image_cl & cr_Image, const Vec2F_st * cap_Points, PatternVtxData_cl * p_VtxData ) = 0;

        virtual MarkerPattern_st &  GetrPattern() = 0;

        /// GbWo[N̈
        EdgeLine_cl         a_MarkerEdge[ e_MarkerEdgeNum ];
        /// @endcond
    };
    /** }[J[oۂɎgpA[N̈NXłB
     */
    struct MarkerDetectWork_cl : public MarkerDetectWork_ac
    {
    public:
        /// @cond DEV
        virtual bool Extract( const Image_cl & cr_Image, const Vec2F_st * cap_Points, PatternVtxData_cl * p_VtxData )
        {
            m_Pattern.Extract( cr_Image, cap_Points, p_VtxData );
            return true;
        }

        virtual MarkerPattern_st &  GetrPattern()    { return m_Pattern; }
        /// @endcond
    private:
        /// }[J[fUCrp[N̈
        MarkerPattern_st    m_Pattern;
    };
    /** }[J[oۂɎgpA[N̈NXłB
     *
     *  }b`ÕTvOɂP_̂ݎgߍł}[J[萸x܂
     */
    struct MarkerDetectWorkFast_cl : public MarkerDetectWork_ac
    {
    public:
        /// @cond DEV
        virtual bool Extract( const Image_cl & cr_Image, const Vec2F_st * cap_Points, PatternVtxData_cl * p_VtxData )
        {
            m_Pattern.ExtractFast( cr_Image, cap_Points, p_VtxData );
            return true;
        }

        virtual MarkerPattern_st &  GetrPattern()    { return m_Pattern; }
        /// @endcond
    private:
        /// }[J[fUCrp[N̈
        MarkerPattern_st    m_Pattern;
    };
    /** }[J[oۂɎgpA[N̈NX̊głłB
     *
     *  }b`Ỏ摜͎̉@قȂ܂B
     */
    struct MarkerDetectWorkStrict_cl : public MarkerDetectWork_ac
    {
    public:
        /// @cond DEV
        virtual bool Extract( const Image_cl & cr_Image, const Vec2F_st * cap_Points, PatternVtxData_cl * p_VtxData )
        {
            m_Pattern.ExtractStrictly( cr_Image, cap_Points, p_VtxData );
            return true;
        }

        virtual MarkerPattern_st &  GetrPattern()    { return m_Pattern; }
        /// @endcond
    private:
        /// }[J[fUCrp[N̈
        MarkerPattern_st    m_Pattern;
    };

    /// @cond DEV

    /** }[J[쐬p̃[N̈NXłB
     */
    template < s32 N >
    class MarkerMakerWork_tc : public MarkerDetectWork_ac
    {
    public:
        MarkerMakerWork_tc()
             : m_UseNum( 0 )
        {}

        /**
         */
        virtual bool Extract( const Image_cl & cr_Image, const Vec2F_st * cap_Points, PatternVtxData_cl * p_VtxData )
        {
            if ( m_UseNum < N )
            {
                ma_Pattern[ m_UseNum ].Extract( cr_Image, cap_Points, p_VtxData );
                maa_Points[ m_UseNum ][ 0 ] = cap_Points[ 0 ];
                maa_Points[ m_UseNum ][ 1 ] = cap_Points[ 1 ];
                maa_Points[ m_UseNum ][ 2 ] = cap_Points[ 2 ];
                maa_Points[ m_UseNum ][ 3 ] = cap_Points[ 3 ];
                m_UseNum++;
            }
            return false;
        }
        /** 
         */
        void PrintPattern( u32 idx )    { ma_Pattern[ idx ].print(); }
        /**
         */
        void Flush() { m_UseNum = 0; }
        /**
         */
        s32 GetPatternNum() { return m_UseNum; }
        /**
         */
        const Vec2F_st *    GetpPoint( u32 idx ) { return maa_Points[ idx ]; }
        /** dummy
         */
        virtual MarkerPattern_st &  GetrPattern()    { return ma_Pattern[ 0 ]; }

        MarkerPattern_st & GetrPattern( u32 idx )    { return ma_Pattern[ idx ]; }
    private:
        s32                     m_UseNum;
        MarkerPattern_st        ma_Pattern[ N ];
        Vec2F_st                maa_Points[ N ][ 4 ];
    };

    /** }[J[쐬p̃[N̈NXłB
     */
    template < s32 N >
    class MarkerMakerWorkStrict_tc : public MarkerDetectWork_ac
    {
    public:
        MarkerMakerWorkStrict_tc()
             : m_UseNum( 0 )
        {}

        /**
         */
        virtual bool Extract( const Image_cl & cr_Image, const Vec2F_st * cap_Points, PatternVtxData_cl * p_VtxData )
        {
            if ( m_UseNum < N )
            {
                ma_Pattern[ m_UseNum ].ExtractStrictly( cr_Image, cap_Points, p_VtxData );
                maa_Points[ m_UseNum ][ 0 ] = cap_Points[ 0 ];
                maa_Points[ m_UseNum ][ 1 ] = cap_Points[ 1 ];
                maa_Points[ m_UseNum ][ 2 ] = cap_Points[ 2 ];
                maa_Points[ m_UseNum ][ 3 ] = cap_Points[ 3 ];
                m_UseNum++;
            }
            return false;
        }
        /** 
         */
        void PrintPattern( u32 idx )    { ma_Pattern[ idx ].print(); }
        /**
         */
        void Flush() { m_UseNum = 0; }
        /**
         */
        s32 GetPatternNum() { return m_UseNum; }
        /**
         */
        const Vec2F_st *    GetpPoint( u32 idx ) { return maa_Points[ idx ]; }
        /** dummy
         */
        virtual MarkerPattern_st &  GetrPattern()    { return ma_Pattern[ 0 ]; }

        MarkerPattern_st & GetrPattern( u32 idx )    { return ma_Pattern[ idx ]; }
    private:
        s32                     m_UseNum;
        MarkerPattern_st        ma_Pattern[ N ];
        Vec2F_st                maa_Points[ N ][ 4 ];
    };

    /// @endcond

    /** 摜}[J[oNXłB
     *
     */
    class MarkerDetector_cl
    {
    public:
        /** }[J[os܂
         *  @param  r_Image oΏۂƂȂ摜(oAjł܂)
         *  @param  r_DB    o}[J[f[^̓f[^x[X
         *  @param  r_Data  o}[J[ۑ̈
         *  @param  r_Work  op[N̈(oAjł܂)
         *  @retval oł}[J[̌Ԃ܂
         */
        static s32 Detect( Image_cl & r_Image, MarkerDatabase_cl & r_DB, MarkerData_ac & r_Data, MarkerDetectWork_ac & r_Work );
    };
}
}
#endif
