/*-----------------------------------------------------------------------*
  Project:  Nintendo Augmented Reality Library.
  File:     narMarkerTrans.h

  Copyright (C)2011-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo and/or its licensed developers
  and are protected by national and international copyright laws. They
  may not be disclosed to third parties or copied or duplicated in any
  form, in whole or in part, without the prior written consent of
  Nintendo.
  The content herein is highly confidential and should be handled
  accordingly.
*-----------------------------------------------------------------------*/

/**
 *  @file
 *  @author Satoru Osako
 */

#ifndef NAR_MARKER_TRANS_H__
#define NAR_MARKER_TRANS_H__

#include "narConfig.h"
#include "narProjection.h"
#include "narFlags.h"

namespace mw { namespace nar
{
    /** }[J[Wn𐄒肷NXłB
     */
    class MarkerTrans_cl
    {
        friend class MarkerTransCore_cl;
    public:
        /**
         */
        MarkerTrans_cl()
             : mcp_Distortion( 0 )
        {}

        /** }[J[̍Wn𐄑܂
         *  @retval łȂꍇfalseԂ܂
         */
        bool Estimate( Marker_cl & r_Marker, const Projection_cl & cr_Proj ) const;

        /** ]sO̍sƕԂ悤ɂ܂
         *  @param b    trueȂԂ܂
         *  @param t    ԌW(0.0:̍sg - 1.0:O̍sg)
         */
        void SetMildRotation( bool b, f32 t = 0.5f )
        {
            m_Flags.Set( b, me_SwitchMildRotation );
            m_InterpolationCoefficient = t;
        }

        /** O̍sƕԂہAkł͖ƔfꍇAfǏ]܂
         *  @param b    trueȂfǏ]
         *  @param t    ԌW(0.0:̍sg - 1.0:O̍sg)
         */
        void SetAccelerationOnMildRotation( bool b, f32 t = 0.25f )
        {
            m_Flags.Set( b, me_SwitchAcceleration );
            m_InterpolationCoefficientOnRapidMovement = t;
        }

        ///	@cond DEV
        /** cȎ␳ݒ
         */
        const Distortion_cl * SetDistortionCorrection( const Distortion_cl * cp )
        {
            const Distortion_cl * cp_Ret = mcp_Distortion;
            mcp_Distortion = cp;
            return cp_Ret;
        }
        ///	@endcond

    private:
        enum Switch_e
        {
            me_SwitchBase,
            NAR_EFLAG( me_SwitchMildRotation ),
            NAR_EFLAG( me_SwitchAcceleration ),
            me_SwitchEnd
        };

        /// XCb`
        Flags_tc< u32, Switch_e >   m_Flags;
        /// ԌW(0.0:current - 1.0:previous)
        f32                         m_InterpolationCoefficient;
        /// }ԌW(0.0:current - 1.0:previous)
        f32                         m_InterpolationCoefficientOnRapidMovement;

        /// cȎ␳
        const Distortion_cl *       mcp_Distortion;
    };
}
}

#endif
