/*-----------------------------------------------------------------------*
  Project:  Nintendo Augmented Reality Library.
  File:     narMesh.h

  Copyright (C)2011-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo and/or its licensed developers
  and are protected by national and international copyright laws. They
  may not be disclosed to third parties or copied or duplicated in any
  form, in whole or in part, without the prior written consent of
  Nintendo.
  The content herein is highly confidential and should be handled
  accordingly.
*-----------------------------------------------------------------------*/
/**
 *  @file
 *  @author Ryota Maruko
 */
#ifndef NAR_MESH_H__
#define NAR_MESH_H__

#include "narPort.h"
#include "narVec2.h"

namespace mw { namespace nar
{
    class Image_cl;

    /** }[J[̃p^[𕪊ӂ̃bV̐łB
     *  NAR̃}[J[oe_MeshDiv15gp܂B
     */
    enum MeshDiv_e
    {
        /// 15
        e_MeshDiv15 = 15,
        /// 31 
        e_MeshDiv31 = 31, 
        /// 63
        e_MeshDiv63 = 63 
    };

    /** p^[rpJ[\
     *  tȂ̂Œ
     *  5bitȂ̂ŁA(0,31)ŗǂ
     */
    struct Color_st
    {
        s8 r,g,b;
    };

    /** bṼJ[\ۃNXłB
     */
    class MeshColorData_ac
    {
    public:
        /** J[擾܂B
         */
        virtual const Color_st &
            GetColor( s32 t, s32 s ) const = 0;

        /** J[擾܂B
         */
        virtual Color_st &
            GetColor( s32 t, s32 s ) = 0;
    };

    /** opbV̊iq_̃J[NXłB
     * @tparam N    bV̈ӓ̊iq_(MeshDiv_e)w肵܂B
     */
    template< MeshDiv_e N >
    class MeshColorData_tc : public MeshColorData_ac
    {
    public:
        /** J[擾܂B
         */
        virtual const Color_st &
            GetColor( s32 t, s32 s ) const { return maa_Data[ t ][ s ]; }

        /** J[擾܂B
         */
        virtual Color_st &
            GetColor( s32 t, s32 s ) { return maa_Data[ t ][ s ]; }
    protected:
        Color_st    maa_Data[ N ][ N ];
        s8          pad[ 5 ];
    };

    /** opbV̊iq_W\ۃNXłB
     */
    class MeshVtxData_ac
    {
    public:
        /** W擾܂B
         */
        virtual const Vec2F_st &
            GetVtx( s32 t, s32 s ) const = 0;

        /** W擾܂B
         */
        virtual Vec2F_st &
            GetVtx( s32 t, s32 s ) = 0;
    };

    /** bV̊iq_̍WNXłB
     * @tparam N    bV̈ӓ̊iq_(MeshDiv_e)w肵܂B
     */
    template< MeshDiv_e N >
    class MeshVtxData_tc : public MeshVtxData_ac
    {
    public:
        /** W擾܂B
         */
        virtual const Vec2F_st &
            GetVtx( s32 t, s32 s ) const { return maa_Data[ t ][ s ]; }

        /** W擾܂B
         */
        virtual Vec2F_st &
            GetVtx( s32 t, s32 s ) { return maa_Data[ t ][ s ]; }
    protected:
        Vec2F_st    maa_Data[ N ][ N ];
    };

    /// @cond DEV

    /** MeshvZp̃[N\ۃNXłB
     */
    class MeshStrictWork_ac
    {
    public:
        virtual Vec2F_st &
            getGrid( s32 nEdge, s32 nDiv ) = 0;

        virtual const Vec2F_st &
            getGrid( s32 nEdge, s32 nDiv ) const = 0;

        virtual Vec2F_st *
            getpGrid( s32 nEdge ) = 0;

        virtual f32
            getST( s32 nST, s32 nDiv ) const = 0;

        virtual f32 *
            getpST( s32 nST ) = 0;

        virtual s32
            getDiv() const = 0;
    };

    /** MeshvZp̃[N\NXłB
     * @tparam N    bV̊iq_w肵܂B
     */
    template< s32 N >
    class MeshStrictWork_tc : public MeshStrictWork_ac
    {
    public:
        virtual Vec2F_st &
            getGrid( s32 nEdge, s32 nDiv ) { return maa_Grid[ nEdge ][ nDiv ]; }

        virtual const Vec2F_st &
            getGrid( s32 nEdge, s32 nDiv ) const { return maa_Grid[ nEdge ][ nDiv ]; }

        virtual Vec2F_st *
            getpGrid( s32 nEdge ) { return maa_Grid[ nEdge ]; }

        virtual f32
            getST( s32 nST, s32 nDiv ) const { return maa_ST[ nST ][ nDiv ]; }

        virtual f32 *
            getpST( s32 nST ) { return maa_ST[ nST ]; }

        virtual s32
            getDiv() const { return N; }

    private:
        Vec2F_st    maa_Grid[ 4 ][ N + 2 ];
        f32         maa_ST[ 2 ][ N + 2 ];
    };

    /** MeshvZNX
     */
    class MeshStrict_cl
    {
    public:
        MeshStrict_cl(  MeshStrictWork_ac & work ) : mr_Work( work )
        {}

        /// OvZ
        void    precalc( const Vec2F_st * cap_Points );

#if NAR_DEBUG__
        /// fobO\
        void    drawInfo( const Image_cl & cr_Image, const Vec2F_st * cap_Points, bool isMarker );
#endif
        /// J[̒o
        void    pickup( const Image_cl & cr_Image, MeshColorData_ac & r_MeshColor );

        /// J[ƍW𒊏o
        void    pickup( const Image_cl & cr_Image, MeshColorData_ac & r_MeshColor, MeshVtxData_ac & r_MeshVtx );

        /// J[̒o
        void    pickupFast( const Image_cl & cr_Image, MeshColorData_ac & r_MeshColor );

        /// J[ƍW𒊏o
        void    pickupFast( const Image_cl & cr_Image, MeshColorData_ac & r_MeshColor, MeshVtxData_ac & r_MeshVtx );

        /// ỏߒŎZoItZbg擾
        s8      getOffset() const { return m_Offset; }
    private:
        Vec2F_st            m_v10;
        Vec2F_st            m_v30;
        Vec2F_st            m_v21;
        Vec2F_st            m_v23;
        MeshStrictWork_ac & mr_Work;
        s8                  m_Offset;
        u8                  pad1[ 3 ];
#if NAR_DEBUG__
        bool                m_isCalcVp;
        bool                m_isDivEq;
        u8                  pad2[ 2 ];
#endif
        void    precalcVp( const Vec2F_st * cap_Points );
        void    precalcApprox( const Vec2F_st * cap_Points, const f32 * cap_L, const f32 * a_Cos );
    };
    /// @endcond
}
}
#endif
