/*-----------------------------------------------------------------------*
  Project:  Nintendo Augmented Reality Library.
  File:     narEdgeLine.h

  Copyright (C)2011-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo and/or its licensed developers
  and are protected by national and international copyright laws. They
  may not be disclosed to third parties or copied or duplicated in any
  form, in whole or in part, without the prior written consent of
  Nintendo.
  The content herein is highly confidential and should be handled
  accordingly.
*-----------------------------------------------------------------------*/

/// @cond DEV
/**
 *  @file
 *  @author Satoru Osako
 */
#ifndef NAR_MESH_LINE_H__
#define NAR_MESH_LINE_H__

#include "narVec2.h"

namespace mw { namespace nar
{
    /**
     *  bVvZp̒NX
     */
    class MeshLine_cl
    {
    public:
        MeshLine_cl()
        {
        }

        /** Jn_o^
         */
        void SetStart( f32 x, f32 y )
        {
            m_Start.Set( x, y );
        }

        /** Jn_o^
         */
        void SetStart( const Vec2F_st & cr )
        {
            SetStart( cr.X(), cr.Y() );
        }

        /** I_o^
         */
        void SetEnd( f32 x, f32 y )
        {
            m_End.Set( x, y );
        }

        /** I_o^
         */
        void SetEnd( const Vec2F_st & cr )
        {
            SetEnd( cr.X(), cr.Y() );
        }

        /** _̎Zo
         */
        bool GetIntersection( Vec2F_st * p_Point, const MeshLine_cl & r_Edge ) const;
        
        f32   GetA() const { return m_A; }
        f32   GetB() const { return m_B; }
        f32   GetC() const { return m_C; }

        /** ̓擾
         */
        s32  GetSquareLength() const
        {
            return GetrStart().GetSquareDistance( GetrEnd() );
        }

        /** startend璼vZ
         */
        bool CalcEquation();

        /**
         *  _ƒ̋߂B
         */
        f32 GetDistance( const Vec2F_st &r ) const;

        /**
         *  _ƒ̋߂B
         */
        f32 GetDistance( const Vec2I_st &r ) const;

    private:
        ///   ax + by + c = 0  a
        f32         m_A;
        ///   ax + by + c = 0  b
        f32         m_B;
        ///   ax + by + c = 0  c
        f32         m_C;

        Vec2F_st    m_Start;
        Vec2F_st    m_End;

        const Vec2F_st & GetrStart() const { return m_Start; }
        const Vec2F_st & GetrEnd()  const { return m_End; }
    };
}
}
#endif
/// @endcond
