/*-----------------------------------------------------------------------*
  Project:  Nintendo Augmented Reality Library.
  File:     narMtx33.h

  Copyright (C)2011-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo and/or its licensed developers
  and are protected by national and international copyright laws. They
  may not be disclosed to third parties or copied or duplicated in any
  form, in whole or in part, without the prior written consent of
  Nintendo.
  The content herein is highly confidential and should be handled
  accordingly.
*-----------------------------------------------------------------------*/

/// @cond DEV
/**
 *  @file
 *  @author Satoru Osako
 */

#ifndef NAR_MTX_33_H__
#define NAR_MTX_33_H__

#include "narPort.h"
#include "narMath.h"

namespace mw { namespace nar
{
    /**
     *  3x3s(En)
     */
    struct Mtx33RF_st
    {
        static const int c_RowNum   = 3;
        static const int c_ColNum   = 3;

        /**
         *
         */
        void MultVec( Vec3F_st & r, const Vec3F_st & cr )
        {
            r.Set( maa_[ 0 ][ 0 ] * cr.X() + maa_[ 0 ][ 1 ] * cr.Y() + maa_[ 0 ][ 2 ] * cr.Z(),
                   maa_[ 1 ][ 0 ] * cr.X() + maa_[ 1 ][ 1 ] * cr.Y() + maa_[ 1 ][ 2 ] * cr.Z(),
                   maa_[ 2 ][ 0 ] * cr.X() + maa_[ 2 ][ 1 ] * cr.Y() + maa_[ 2 ][ 2 ] * cr.Z() );
        }
        /**
         *
         */
        void MakeRotateXYZ( u16 xaxis, u16 yaxis, u16 zaxis )
        {
#ifdef NAR_CTR__
            f32 cosx = nn::math::CosIdx( xaxis );
            f32 sinx = nn::math::SinIdx( xaxis );
            f32 cosy = nn::math::CosIdx( yaxis );
            f32 siny = nn::math::SinIdx( yaxis );
            f32 cosz = nn::math::CosIdx( zaxis );
            f32 sinz = nn::math::SinIdx( zaxis );
#else
            fixme
#endif
            maa_[ 0 ][ 0 ]  =  cosy * cosz;
            maa_[ 0 ][ 1 ]  = -cosy * sinz;
            maa_[ 0 ][ 2 ]  =  siny;

            maa_[ 1 ][ 0 ]  =  cosx * sinz + sinx * siny * cosz;
            maa_[ 1 ][ 1 ]  =  cosx * cosz - sinx * siny * sinz;
            maa_[ 1 ][ 2 ]  = -sinx * cosy;

            maa_[ 2 ][ 0 ]  =  sinx * sinz - cosx * siny * cosz;
            maa_[ 2 ][ 1 ]  =  sinx * cosz + cosx * siny * sinz;
            maa_[ 2 ][ 2 ]  =  cosx * cosy;
        }

        union
        {
            f32 ma_[ c_RowNum * c_ColNum ];
            f32 maa_[ c_RowNum ][ c_ColNum ];
        };
    };
}
}
#endif
/// @endcond
