/*-----------------------------------------------------------------------*
  Project:  Nintendo Augmented Reality Library.
  File:     narProjection.h

  Copyright (C)2011-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo and/or its licensed developers
  and are protected by national and international copyright laws. They
  may not be disclosed to third parties or copied or duplicated in any
  form, in whole or in part, without the prior written consent of
  Nintendo.
  The content herein is highly confidential and should be handled
  accordingly.
*-----------------------------------------------------------------------*/

/**
 *  @file
 *  @author Satoru Osako
 */

#ifndef NAR_PROJECTION_H__
#define NAR_PROJECTION_H__

#include "narPort.h"

#ifdef NAR_CTR__
namespace nn {
    namespace math {
        struct MTX44;
    }
}
#endif

namespace mw { namespace nar
{
    struct Vec2F_st ;
    struct Vec3F_st ;
    class Distortion_cl;
    /** ˉeϊsNXłB
     */
    class Projection_cl
    {
    public:
        Projection_cl()
        {}

        /**
         *  @param w    摜̕
         *  @param h    摜̍
         *  @param near nearʂ܂ł̋
         *  @param far  farʂ܂ł̋
         *  @param aov  Ίpp(degrees)
         */
        Projection_cl( f32 w, f32 h, f32 near, f32 far, f32 aov = 66.0 )
        {
            Set( w, h, near, far, aov );
        }

        /** ˉesݒ肵܂B
         *  @param w    摜̕
         *  @param h    摜̍
         *  @param near nearʂ܂ł̋
         *  @param far  farʂ܂ł̋
         *  @param aov  Ίpp(degrees)
         */
        void Set( f32 w, f32 h, f32 near, f32 far, f32 aov = 66.0 );
#ifdef NAR_TWL__
        /** Twls擾
         */
        void GetMtxFx44( MtxFx44 & r_Mtx ) const ;
#endif
#ifdef NAR_CTR__
        /** ˉep̍s擾܂
         */
        void GetMTX44( nn::math::MTX44 & r_Mtx ) const;

        /** g~OˉesԂ܂
         *  @param[out] r_Mtx
         *  @param[in]  l, r, b, t  g~OȂ̂́Ax,yW͈̔͂[-1,1]Ƃۂ́Ag~Oʒuw肵܂B
         *  @param      n       nearNbsOʂ܂ł̋
         *  @param      f       farNbsOʂ܂ł̋
         *  @param      aov     Angle of View(degree)
         */
        void GetTrimmed( nn::math::MTX44 & r_Mtx, f32 l, f32 r, f32 b, f32 t, f32 n, f32 f, f32 aov = 66.f ) const;
#endif
        /// @cond DEV
        
        /** 摜ʏ̓_Ɏˉereferenceʏ̓_擾܂
         *  @param[out] r_Vec3 referenceʏ̓_
         *  @param[in]  r_Vec2 摜ʏ̓_
         */
        void GetVtxOnReferencePlane( Vec3F_st & r_Vec3, const Vec2F_st & r_Vec2 ) const;

        /** cȎ␳ĉ摜ʏ̓_Ɏˉereferenceʏ̓_擾܂B
         *  @param[out] r_Vec3  referenceʏ̓_
         *  @param[in]  r_Vec2  摜ʏ̓_
         *  @param[in]  cr_Dist distotionNX
         */
        void GetVtxOnReferencePlane( Vec3F_st & r_Vec3, const Vec2F_st & cr_Vec2, const Distortion_cl & cr_Dist ) const;
        /// @endcond
    private:
        f32     m_00;
        f32     m_11;
        f32     m_22;
        f32     m_32;
        f32     m_HalfWidth;
        f32     m_HalfHeight;
    };
}
}

#endif
