/*-----------------------------------------------------------------------*
  Project:  Nintendo Augmented Reality Library.
  File:     narTransformation.h

  Copyright (C)2011-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo and/or its licensed developers
  and are protected by national and international copyright laws. They
  may not be disclosed to third parties or copied or duplicated in any
  form, in whole or in part, without the prior written consent of
  Nintendo.
  The content herein is highly confidential and should be handled
  accordingly.
*-----------------------------------------------------------------------*/

/**
 *  @file
 *  @author Satoru Osako
 */
#ifndef NAR_TRANSFORMATION_H__
#define NAR_TRANSFORMATION_H__

#include "narVec3.h"

#ifdef NAR_CTR__
namespace nn {
    namespace math {
        struct MTX44;
    }
}
#endif

namespace mw { namespace nar
{
    struct Mtx33RF_st;
    /** }[J[WnJWnɕϊsNXłB
     */
    class Transformation_cl
    {
    public:
#ifdef NAR_TWL__
        /** }[J[WnJWnɕϊs擾܂
         */
        void GetMtxFx43( MtxFx43 & r_Mtx ) const;
#endif
#ifdef NAR_CTR__
        /** }[J[WnJWnɕϊs擾܂
         */
        void GetMTX34( nn::math::MTX34 & r_Mtx ) const;
#endif

        /// @cond DEV
        /** XxNg擾
         */
        const Vec3F_st & GetrXAxis() const { return ma_Vec[ me_X ]; }
        
        /** YxNg擾
         */
        const Vec3F_st & GetrYAxis() const { return ma_Vec[ me_Y ]; }
        
        /** ZxNg擾
         */
        const Vec3F_st & GetrZAxis() const { return ma_Vec[ me_Z ]; }
        
        /** ixNg擾
         */
        const Vec3F_st & GetrTrans() const { return ma_Vec[ me_Trans ]; }

        /** ]s̃Zbg
         */
        void SetRot( const Vec3F_st & cr_X, const Vec3F_st & cr_Y, const Vec3F_st & cr_Z )
        {
            ma_Vec[ me_X ] = cr_X;
            ma_Vec[ me_Y ] = cr_Y;
            ma_Vec[ me_Z ] = cr_Z;
        }

        /** sړxNgZbg
         */
        void SetTrans( const Vec3F_st & cr_Trans )
        {
            ma_Vec[ me_Trans ] = cr_Trans;
        }

        /** Qƕʂ֕ϊ
         */
        void TransformToRefPlane( Vec3F_st & r, const Vec3F_st & cr ) const;

        /** s̘A
         */
        void Concatenate( Mtx33RF_st & r );
        /// @endcond
    private:
        enum
        {
            me_X, me_Y, me_Z, me_Trans, me_Max
        };

        Vec3F_st    ma_Vec[ me_Max ];
    };
}
}

#endif
