/*-----------------------------------------------------------------------*
  Project:  Nintendo Augmented Reality Library.
  File:     narVec3.h

  Copyright (C)2011-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo and/or its licensed developers
  and are protected by national and international copyright laws. They
  may not be disclosed to third parties or copied or duplicated in any
  form, in whole or in part, without the prior written consent of
  Nintendo.
  The content herein is highly confidential and should be handled
  accordingly.
*-----------------------------------------------------------------------*/

/// @cond DEV
/**
 *  @file
 *  @author Satoru Osako
 */

#ifndef NAR_VEC_3_H__
#define NAR_VEC_3_H__

#include "narPort.h"
#include "narMath.h"

#ifdef NAR_CTR__
namespace nn {
    namespace math {
        struct VEC3;
    }
}
#endif

namespace mw { namespace nar
{
    /**
     *  3xNg\
     */
    struct Vec3F_st
    {
        Vec3F_st()
        {}

        Vec3F_st( f32 x, f32 y, f32 z )
        {
            Set( x, y, z );
        }

        Vec3F_st( f32 val )
        {
            SetAll( val );
        }

        void Set( f32 x, f32 y, f32 z )
        {
            X() = x; Y() = y; Z() = z;
        }

        void SetAll( f32 val )
        {
            Set( val, val, val );
        }

        f32 & X() { return ma_[ me_X ]; }
        f32 & Y() { return ma_[ me_Y ]; }
        f32 & Z() { return ma_[ me_Z ]; }
        const f32 & X() const { return ma_[ me_X ]; }
        const f32 & Y() const { return ma_[ me_Y ]; }
        const f32 & Z() const { return ma_[ me_Z ]; }


        /** a
         */
        Vec3F_st & Add( const Vec3F_st & cr )
        {
            X() += cr.X();
            Y() += cr.Y();
            Z() += cr.Z();
            return *this;
        }

        /** 
         */
        Vec3F_st & Sub( const Vec3F_st & cr )
        {
            X() -= cr.X();
            Y() -= cr.Y();
            Z() -= cr.Z();
            return *this;
        }

        /** XJ[{
         */
        Vec3F_st & Scale( f32 a )
        {
            X() *= a;
            Y() *= a;
            Z() *= a;
            return *this;
        }

        /** XJ[{
         */
        Vec3F_st operator*( f32 a ) const
        {
            Vec3F_st ret( *this );
            return ret.Scale( a );
        }

        /** XJ[
         */
        Vec3F_st & Div( f32 d )
        {
            if ( d != 0.f )
            {
                Scale( 1.f / d );
            }
            return * this;
        }
        
        /** hbg
         */
        f32 Dot( const Vec3F_st & cr ) const
        {
            return X() * cr.X() + Y() * cr.Y() + Z() * cr.Z();
        }

        /** NX
         */
        Vec3F_st & Cross( const Vec3F_st & cr_A, const Vec3F_st & cr_B )
        {
            f32 ax = cr_A.X();
            f32 ay = cr_A.Y();
            f32 az = cr_A.Z();
            f32 bx = cr_B.X();
            f32 by = cr_B.Y();
            f32 bz = cr_B.Z();

            X() = ay * bz - az * by;
            Y() = az * bx - ax * bz;
            Z() = ax * by - ay * bx;
            return *this;
        }

        /** K
         */
        bool Normalize()
        {
            f32 L = GetLength();
            if ( L < c_Epsilon ) return false;
            Scale( static_cast< f32 >( 1.0 ) / L );
            return true;
        }

        /** ̓߂
         */
        f32 GetSquareLength() const
        {
            return Pow2( X() ) + Pow2( Y() ) + Pow2( Z() );
        }

        /** ߂
         */
        f32 GetLength() const
        {
            return Sqrt( GetSquareLength() );
        }

        /** ̓_Ƃ̋̓
         */
        f32 GetSquareDistance( const Vec3F_st & cr_Vec ) const
        {
            return Pow2( cr_Vec.X() - X() ) + Pow2( cr_Vec.Y() - Y() ) + Pow2( cr_Vec.Z() - Z() );
        }

        /** ̓_Ƃ̋
         */
        f32 GetDistance( const Vec3F_st & cr_Vec ) const
        {
            return Sqrt( GetSquareDistance( cr_Vec ) );
        }

        enum
        {
            me_X, me_Y, me_Z, me_ArrayNum
        };

        f32 ma_[ me_ArrayNum ];

#ifdef NAR_CTR__
        void CopyTo( nn::math::VEC3 &r ) const
        {
            r.x = X(); r.y = Y(); r.z = Z();
        }
#endif
    };
}
}
#endif
/// @endcond
