/*-----------------------------------------------------------------------*
  Project:  Nintendo Augmented Reality Library.
  File:     sampledemos\common\narDemoCamera.cpp

  Copyright (C)2011 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo and/or its licensed developers
  and are protected by national and international copyright laws. They
  may not be disclosed to third parties or copied or duplicated in any
  form, in whole or in part, without the prior written consent of
  Nintendo.
  The content herein is highly confidential and should be handled
  accordingly.
*-----------------------------------------------------------------------*/

/**
 *  @file
 *  @author Satoru Osako
 */

#include <nn.h>
#include <nn/camera.h>

#include "narDemoConfig.h"
#include "narDemoCamera.h"

namespace
{
    const size_t    c_CameraThreadStackSize = 0x1000;
}
//=====================================================================
void CameraProcess_cl::StartThread()
{
    m_Thread.StartUsingAutoStack( hook,
                                  this,
                                  c_CameraThreadStackSize,
                                  nn::os::Thread::GetCurrentPriority() - 3 );
}
//=====================================================================
void CameraProcess_cl::Initialize( nn::camera::Port port, u32 transLine, nn::fnd::ExpHeap * p_Heap )
{
    m_Port  = port;
    mp_Heap = p_Heap;

    //
    m_CriticalSection.Initialize();

    ma_Events[ me_CameraTransferEvent ].Initialize( false );
    ma_Events[ me_CameraErrorEvent    ].Initialize( false );
    ma_Events[ me_CameraEndEvent      ].Initialize( false );
    
    //
    nn::camera::SetTransferLines( port, transLine, IMG_WIDTH, IMG_HEIGHT );

    // G[荞݃nh̎擾
    nn::camera::GetBufferErrorInterruptEvent( &ma_Events[ me_CameraErrorEvent ], port );

    // g~O̐ݒ
    nn::camera::SetTrimming( port, false );

    //  obt@
    maap_YuvBuffer  = reinterpret_cast< u16 * >( mp_Heap->Allocate( me_YuvBufferNum * IMG_WIDTH * IMG_HEIGHT * YUV_BYTES, 64 ) );
}
//=====================================================================
void CameraProcess_cl::LockImage()
{
    //  NeBJZNVɓ
    nn::os::CriticalSection::ScopedLock sl( m_CriticalSection );
    m_Locked    = m_Captured;
}
//=====================================================================
void CameraProcess_cl::setReceiving()
{
    nn::camera::SetReceiving( &ma_Events[ me_CameraTransferEvent ],
                              maap_YuvBuffer + m_Capturing * IMG_PIXEL_NUM,
                              m_Port,
                              nn::camera::GetFrameBytes( IMG_WIDTH, IMG_HEIGHT ),
                              nn::camera::GetTransferBytes( m_Port ) );
}
//=====================================================================
void CameraProcess_cl::process()
{
    nn::os::Event * pa_Events[] =
    {
        &ma_Events[ me_CameraTransferEvent ],        &ma_Events[ me_CameraErrorEvent ],
        &ma_Events[ me_CameraEndEvent ],
    };

    setReceiving();

    nn::camera::ClearBuffer( m_Port );
    nn::camera::StartCapture( m_Port );

    while ( true )
    {
        nn::fnd::TimeSpan start = nn::os::Tick::GetSystemCurrent().ToTimeSpan();

        //  ]AG[҂
        s32 id = nn::os::WaitObject::WaitAny( reinterpret_cast< nn::os::WaitObject ** >( pa_Events ), sizeof( pa_Events ) / sizeof( pa_Events[ 0 ] ) );

        nn::fnd::TimeSpan end   = nn::os::Tick::GetSystemCurrent().ToTimeSpan();

        switch ( id )
        {
        case me_CameraTransferEvent:
            {
                nn::os::CriticalSection::ScopedLock sl( m_CriticalSection );

                m_Captured  = m_Capturing;
                do
                {
                    if ( ++m_Capturing >= me_YuvBufferNum ) m_Capturing = 0;
                }
                while ( m_Capturing == m_Locked );  //  ̃obt@͎gpȂ
            }
            ma_ShootingTime[ m_Captured ]   = end;
            ma_TransferTime[ m_Captured ]   = end - start;
            m_CapturedEvent.Signal();
            setReceiving();
            break;
        case me_CameraErrorEvent:
            //  Lv`ĊJ
            nn::camera::StopCapture( m_Port );
            nn::camera::ClearBuffer( m_Port );
            setReceiving();
            nn::camera::StartCapture( m_Port );
            break;
        case me_CameraEndEvent:
            return;
        }
    }
}
//=====================================================================
void CameraProcess_cl::Finalize()
{
    ma_Events[ me_CameraEndEvent ].Signal();
    m_Thread.Join();
    m_Thread.Finalize();

    mp_Heap->Free( reinterpret_cast< void * >( maap_YuvBuffer ) );
}
