/*-----------------------------------------------------------------------*
  Project:  Nintendo Augmented Reality Library.
  File:     sampledemos\common\narDemoY2r.cpp

  Copyright (C)2011 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo and/or its licensed developers
  and are protected by national and international copyright laws. They
  may not be disclosed to third parties or copied or duplicated in any
  form, in whole or in part, without the prior written consent of
  Nintendo.
  The content herein is highly confidential and should be handled
  accordingly.
*-----------------------------------------------------------------------*/

/**
 *  @file
 *  @author Satoru Osako
 */
#include <nn.h>
#include <nn/camera.h>
#include <nn/y2r.h>

#include "narDemoConfig.h"
#include "narDemoY2r.h"


//=====================================================================
void Y2RProcess_cl::Initialize()
{
    // Y2R 
    nn::y2r::Initialize();

    // Y2R̐ݒ

    // ̓tH[}bg̕ύX
    nn::y2r::SetInputFormat( nn::y2r::INPUT_YUV422_BATCH );

    // o̓tH[}bg̕ύX
    nn::y2r::SetOutputFormat( nn::y2r::OUTPUT_RGB_16_555 );

    // o̓f[^̉]px̕ύX
    nn::y2r::SetRotation( nn::y2r::ROTATION_NONE );

    // o̓f[^̕я̕ύX
    nn::y2r::SetBlockAlignment( nn::y2r::BLOCK_8_BY_8 );

    // 荞݃nh̎擾
    nn::y2r::GetTransferEndEvent( &m_TransEndEvent );

    // ̓f[^̃TCYw
    nn::y2r::SetInputLineWidth( IMG_WIDTH );
    nn::y2r::SetInputLines( IMG_HEIGHT );

    // ϊW̐ݒ
    nn::y2r::StandardCoefficient c;
    nn::camera::GetSuitableY2rStandardCoefficient( &c );
    nn::y2r::SetStandardCoefficient( c );

    // At@l̐ݒ( ̃fł͎gpȂ )
    nn::y2r::SetAlpha( 0xFF );

    //  XbhX^[g
    m_Thread.StartUsingAutoStack( hook, this, 4096, nn::os::Thread::GetCurrentPriority() - 2 );
}
//=====================================================================
void Y2RProcess_cl::Push( Y2RJob_cl & r )
{
    if ( m_EndRequest ) return;

    Y2RJob_cl **    pp_Job  = &mp_WorkingJob;

    while ( *pp_Job )
    {
        pp_Job = &( ( *pp_Job )->mp_Next );
    }
    *pp_Job = &r;
}
//=====================================================================
void Y2RProcess_cl::process()
{
    while ( true )
    {
        //  L[҂
        m_NewJobEvent.Wait();

        //  INGXg
        if ( m_EndRequest )
            break;

        do
        {
            //  Zbg
            nn::y2r::SetSendingYuv( mp_WorkingJob->GetpYuv(), IMG_BYTES, IMG_WIDTH * YUV_BYTES );
            nn::y2r::SetReceiving( mp_WorkingJob->GetpRgb(), nn::y2r::GetOutputImageSize( IMG_WIDTH, IMG_HEIGHT, nn::y2r::OUTPUT_RGB_16_555 ), IMG_WIDTH * RGB_BYTES * 8 );
            nn::y2r::StartConversion();

            //  I҂
            //	Y2RnOAbv΍Ƃ20ms̃^CAEgĂ
            bool finish = m_TransEndEvent.Wait( nn::fnd::TimeSpan::FromMilliSeconds( 20 ) );

            if ( finish )
            {
                mp_WorkingJob = mp_WorkingJob->mp_Next;
            }
            else
            {
                //	IĂȂ΋I
                nn::y2r::StopConversion();
            }
        }
        while ( mp_WorkingJob );

        m_JobDoneEvent.Signal();
    }
}
//=====================================================================
void Y2RProcess_cl::Finalize()
{
    m_EndRequest    = true;
    m_NewJobEvent.Signal();

    m_Thread.Join();
    m_Thread.Finalize();

    nn::y2r::Finalize();
}

