﻿/*---------------------------------------------------------------------------*
  Project:  Horizon

  Copyright (C)Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 58466 $
 *---------------------------------------------------------------------------*/

#include "GuiButton.h"

namespace scene
{

// コンストラクタです。
Button::Button(u32 id, s32 x, s32 y, u32 width, u32 height, const wchar_t* pText, void* pExtraData, f32 fontScale)
    : ControlBase(CONTROL_TYPE_BUTTON, id, x, y, width, height, pText, pExtraData, fontScale)
{
    m_textAlign   = TEXT_ALIGN_CENTER;
    m_borderWidth = 2.0f;
    m_state       = CONTROL_STATE_DEFAULT;

    // 枠線の色
    m_borderColors[CONTROL_STATE_DEFAULT]   = nn::util::Color8(192, 192, 192, 255);
    m_borderColors[CONTROL_STATE_DISABLED]  = nn::util::Color8( 32,  32,  32, 255);
    m_borderColors[CONTROL_STATE_SELECTING] = nn::util::Color8( 45, 116, 174, 255);

    // 文字の色
    m_textColors[CONTROL_STATE_DEFAULT]   = nn::util::Color8(192, 192, 192, 255);
    m_textColors[CONTROL_STATE_DISABLED]  = nn::util::Color8( 32,  32,  32, 255);
    m_textColors[CONTROL_STATE_SELECTING] = nn::util::Color8( 45, 116, 174, 255);
}

// デストラクタです。
Button::~Button()
{
}

// ペンで押された時の処理
void Button::OnPenDown()
{
    m_state = CONTROL_STATE_SELECTING;
}

// ペンが離された時の処理
void Button::OnPenUp(bool isIn)
{
    (void)isIn;

    m_state = CONTROL_STATE_DEFAULT;
}

// ペンでスライドされた時の処理
void Button::OnPenSlide(bool isIn)
{
    m_state = (isIn ? CONTROL_STATE_SELECTING : CONTROL_STATE_DEFAULT);
}

// ペンでタッチされた時の処理
void Button::OnPenTouch()
{
}

// 描画時の処理
void Button::OnDraw()
{
    // レンダラ
    demo::RenderSystemDrawing* pRenderSystem = GetRenderSystem();
    // テキストライター
    TextWriter* pTextWriter = GetTextWriter();

    // 枠線
    if (m_borderWidth >= 0.0f)
    {
        nn::util::FloatColor color = m_borderColors[m_state];

        // 箱を描画する
        pRenderSystem->SetColor(color.r, color.g, color.b, color.a);
        scene::util::DrawBox(m_x, m_y, m_width, m_height, m_borderWidth);
    }

    // テキスト
    if (m_textLength > 0)
    {
        nn::util::Color8 color = m_textColors[m_state];
        // スケール設定
        pTextWriter->SetFontScale(m_TextScale);

        // テキスト描画位置
        s32 tx = CalculateTextX(2, 0);
        s32 ty = CalculateTextY(0);

        // テキストを描画する
        pTextWriter->SetTextColor(color.r, color.g, color.b);
        pTextWriter->PutString(tx, ty, m_pText);
    }
}

}
