/*-----------------------------------------------------------------------*
  Project:  Nintendo Augmented Reality Library.
  File:     sampledemos\simple\main.cpp

  Copyright (C)2011 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo and/or its licensed developers
  and are protected by national and international copyright laws. They
  may not be disclosed to third parties or copied or duplicated in any
  form, in whole or in part, without the prior written consent of
  Nintendo.
  The content herein is highly confidential and should be handled
  accordingly.
*-----------------------------------------------------------------------*/

/**
 *  @file
 *  CTRARgp̊ȒPȃTvłB
 *  @author Satoru Osako
 */

#include <nn.h>

#include <nn/fnd.h>

#include <nn/gx.h>
#include <nn/math.h>
#include <nn/fs.h>

#include <nn/hid.h>
#include <nn/camera.h>
#include <nn/y2r.h>

#include "nar.h"

#include "cube.h"

#include "hatenaMarkerTemplate.h"

#include "narDemoConfig.h"
#include "narDemoCamera.h"
#include "narDemoPhoto.h"
#include "narDemoY2r.h"

#include "demo.h"

#define NEAR        (1)
#define FAR         (5000)

#define MARKER_LIMIT_NUM (10)

// }[J[̓̉ 0:͂Ȃ 1:͂
#define NAR_USE_MAKER_INSIDE__  (1)

namespace
{
    //  摜p[N
    u32 sa_NarImageWork[ NAR_IMAGE_WORK_SIZE_4TH( IMG_WIDTH, IMG_HEIGHT ) ] ;

    //  }[J[op[N
    mw::nar::MarkerDetectWork_cl                s_DetectWork;

    //  }[J[op[NivZ͂xmȌoj
    mw::nar::MarkerDetectWorkStrict_cl          s_DetectWorkStrict;

#if NAR_USE_MAKER_INSIDE__

    // }[J[̓̔Cӂ̏ꏊɉ͂ꍇ̕imw::nar::MeshDiv_e15,31,63̕Ij
    const mw::nar::MeshDiv_e                                c_MarkerInnerDiv  = mw::nar::e_MeshDiv15;

    // bV̍Wێ}[J[op̈
    mw::nar::MarkerAnalizableData_tc< MARKER_LIMIT_NUM >    s_MarkerData;
    const int                                               c_auto_stack_size = 0x6000;

#else

    //  ʏ̃}[J[op̈
    mw::nar::MarkerData_tc< MARKER_LIMIT_NUM >              s_MarkerData;
    const int                                               c_auto_stack_size = 0x2000;

#endif

    const int                           c_gl_memory_size = 0x800000;

    demo::RenderSystemDrawing           s_RenderSystem;

    nn::fnd::ExpHeap                    s_AppHeap;
    uptr                                s_HeapForGx;

    mw::nar::MarkerTrans_cl             s_NarTrans;
    mw::nar::Projection_cl              s_Projection;
    
    bool    s_ViewInfo = false;

    /* program id */
    GLuint s_PgID;

    /* shader id */
    GLuint s_ShID;

    //=====================================================================
    /**
     *  JW[
     */
    class CameraModule_cl
    {
    public:
        void    Initialize();
        /** J摜̂҂
         */
        bool    Wait();

        void    Finalize();
        
        CameraProcess_cl &  GetrCameraProcess() { return m_Process; }
    private:
        CameraProcess_cl    m_Process;
        nn::os::Event       m_EndEvent;
        u32                 pad;
    };
    //=====================================================================
    void CameraModule_cl::Initialize()
    {
        // CAMERA 
        nn::camera::Initialize();

        // obt@ɂ߂郉Cw
        u32 transferLine = nn::camera::GetMaxLines( IMG_WIDTH, IMG_HEIGHT );

        if ( IMG_WIDTH < 16 ) // GetMaxLines͍ő僉C擾֐Ȃ̂ŁAX̃Cȉ̏ꍇ蓾B
        {
            transferLine = IMG_HEIGHT;
        }

        //  Nx
        nn::camera::SetSharpness( nn::camera::SELECT_OUT1, 3 );

        //  Io
        nn::camera::SetExposure( nn::camera::SELECT_OUT1, 3 );

        //  t[[g̕ύX
        nn::camera::SetFrameRate( nn::camera::SELECT_OUT1, nn::camera::FRAME_RATE_30 );

        //  摜TCY̕ύX
        nn::camera::SetSize( nn::camera::SELECT_OUT1, CAMERA_SIZE, nn::camera::CONTEXT_A );

        //  JNԂɂ
        nn::camera::Activate( nn::camera::SELECT_OUT1 );

        //  JvZXN
        m_Process.Initialize( nn::camera::PORT_CAM1, transferLine, &s_AppHeap );

        //  IpCxg
        m_EndEvent.Initialize( false );

        //  JvZXX^[g
        m_Process.StartThread();
    }
    //=====================================================================
    bool CameraModule_cl::Wait()
    {
        nn::os::WaitObject *    pa_WaitObjects[ 2 ];
        pa_WaitObjects[ 0 ] = m_Process.GetpCapturedEvent();
        pa_WaitObjects[ 1 ] = &m_EndEvent;

        s32 id = nn::os::WaitObject::WaitAny( pa_WaitObjects, 2 );

        if ( id == 1 )
            return false;
        
        //  ŐṼLv`ʂbN
        m_Process.LockImage();
        return true;
    }
    //=====================================================================
    void CameraModule_cl::Finalize()
    {
        m_Process.Finalize();
        m_EndEvent.Signal();

        nn::camera::Finalize();
    }

    CameraModule_cl s_CameraModule;

    //=====================================================================
    /**
     *  ARvZX
     */
    class ARProcess_cl
    {
    public:
        ARProcess_cl()
             : m_Displaying( 0 ), m_Processing( 0 ), m_Displayable( false ), m_IsWorkStrict( false )
        {}

        void    Initialize();
        void    Finalize();
        bool    IsDisplayable() { return m_Displayable; }
        void    DisplayDone()  { m_Displayable = false; }
        void    SwitchDetectWork() { m_IsWorkStrict = !m_IsWorkStrict; }
        bool    IsDetectedWorkStrict() { return m_IsWorkStrict; }

        Photo_cl &      GetrDisplayablePhoto()          { return ma_Photo[ m_Displaying ]; }
        Photo_cl &      GetrDisplayableInfo()           { return ma_Info[ m_Displaying ]; }
        mw::nar::MarkerList_ac * GetpDisplayableMarkerList()     { return mpa_MarkerLists[ m_Displaying ]; }
        nn::fnd::TimeSpan & GetrDisplayableTransferTime()   { return ma_TransferTime[ m_Displaying ]; }
        nn::fnd::TimeSpan & GetrDisplayableARTime()         { return ma_ARTime[ m_Displaying ]; }
    private:
        s8          m_Displaying;
        s8          m_Processing;
        bool        m_Displayable;
        bool        m_IsWorkStrict;

        Photo_cl                    ma_Photo[ 2 ];
        Photo_cl                    ma_Info[ 2 ];
        mw::nar::MarkerList_ac *    mpa_MarkerLists[ 2 ];
        nn::os::Thread              m_Thread;
        u32                         pad;
        nn::fnd::TimeSpan           ma_TransferTime[ 2 ];
        nn::fnd::TimeSpan           ma_ARTime[ 2 ];

        static void hook( ARProcess_cl * p )
        {
            p->process();
        }
        void    process();
    };
    //=====================================================================
    void ARProcess_cl::Initialize()
    {
        for ( int idx = 0; idx < 2; idx++ )
        {
            ma_Photo[ idx ].Initialize( &s_RenderSystem, &s_AppHeap );
            ma_Info [ idx ].Initialize( &s_RenderSystem, &s_AppHeap );
        }

        //  CXbhႢ
        m_Thread.StartUsingAutoStack( hook, this, c_auto_stack_size, nn::os::Thread::GetCurrentPriority() + 2 );
    }
    //=====================================================================
    void ARProcess_cl::process()
    {
        mw::nar::MarkerDatabase_cl   narDatabase;
        mw::nar::Image_cl            narImage( IMG_WIDTH, IMG_HEIGHT, sa_NarImageWork ) ;

        // neĩ}[J[ev[go^
        narDatabase.Register( g_HatenaMarker );

        while ( true )
        {
            //  J摜̂҂
            if ( ! s_CameraModule.Wait() )
                break;

            //  Y2Rϊsčs

            //  Y2RWuo^
            Y2RJob_cl   job;

            CameraProcess_cl & r_Process = s_CameraModule.GetrCameraProcess();

            job.Set( r_Process.GetpLockedYuv(), ma_Photo[ m_Processing ].GetpImage() );
            Y2RProcess_cl::ss().Push( job );

            //  Y2RϊJn
            Y2RProcess_cl::ss().Start();

            //  AR}[J[o
            nn::fnd::TimeSpan   ARBegin = nn::os::Tick::GetSystemCurrent().ToTimeSpan();

            narImage.SetImage( r_Process.GetpLockedYuv() );
            narImage.SetDebugImage( ma_Info[ m_Processing ].GetpImage() );

            //  }[J[o
            int markerNum = 0;
            if( m_IsWorkStrict )
            {
                markerNum = mw::nar::MarkerDetector_cl::Detect( narImage, narDatabase, s_MarkerData, s_DetectWorkStrict );
            }
            else
            {
                markerNum = mw::nar::MarkerDetector_cl::Detect( narImage, narDatabase, s_MarkerData, s_DetectWork );
            }

            //  AR}[J[W菈
            for ( int idx = 0 ; idx < markerNum ; idx++ )
            {
                mw::nar::Marker_cl * p_Marker = s_MarkerData.GetrDetectingMarkerList().GetpMarker( idx ) ;

                //  k}
                p_Marker->SuppressShake( 4.f, 16.f );

                if ( ! p_Marker->IsEstimated() )
                {
                    // }[J[Wn̐
                    s_NarTrans.Estimate( *p_Marker, s_Projection );
                }

#if NAR_USE_MAKER_INSIDE__
                // XɃ}[J[̍P^Sӂɕĕt擾                
                s32 dir = p_Marker->GetDir();
                const mw::nar::PatternVtxData_cl * p_MarkerVertices = p_Marker->GetpPatternVtxData();

                if( p_MarkerVertices )
                {
                    // o_獶P^Sӂ擾
                    // }[J[15x15ŕ
                    //      s
                    //  P0-----P3
                    //   |     |
                    // t |     | 
                    //   |     |
                    //  P1-----P2
                    //      
                    mw::nar::Vec2F_st a_Points[ 4 ];                                
                    a_Points[ 0 ].Set( p_MarkerVertices->GetDirVtx( 7, 0, dir ) );
                    a_Points[ 1 ].Set( p_MarkerVertices->GetDirVtx( 14, 0, dir ) );
                    a_Points[ 2 ].Set( p_MarkerVertices->GetDirVtx( 14, 7, dir ) );
                    a_Points[ 3 ].Set( p_MarkerVertices->GetDirVtx( 7, 7, dir ) );

                    // S_̓𕪊Č_̐F擾
                    mw::nar::MeshColorData_tc< c_MarkerInnerDiv >  colorData;
                    narImage.PickupMeshData< c_MarkerInnerDiv >( a_Points, colorData );

                    // S_̓𕪊Č_̐FƍW擾
                    //mw::nar::MeshColorData_tc< c_MarkerInnerDiv >  colorData;
                    //mw::nar::MeshVtxData_tc< c_MarkerInnerDiv >    vtxData;
                    //narImage.PickupMeshData< c_MarkerInnerDiv >( a_Points, colorData, vtxData );

                    // 擾̏gĕʂ̉͂s
                    //for( int t=0; t < c_MarkerInnerDiv; t++ )
                    //for( int s=0; s < c_MarkerInnerDiv; s++ )
                    //{
                    //    mw::nar::Color_st & c = colorData.GetColor( t, s );
                    //    
                    //}

                }
#endif
            }

            ma_ARTime[ m_Processing ]   = nn::os::Tick::GetSystemCurrent().ToTimeSpan() - ARBegin;

            //  ]ԕۑ
            ma_TransferTime[ m_Processing ] = r_Process.GetrTransferTimeOfLockedImage();
            //  oʂւ̃|C^ێ
            mpa_MarkerLists[ m_Processing ] = &( s_MarkerData.GetrDetectingMarkerList() );

            //  Y2Rϊ҂
            Y2RProcess_cl::ss().WaitJobFinish();

            m_Displaying    = m_Processing;
            //  ̃vZXCfbNX
            m_Processing ^= 0x1;

            m_Displayable   = true;
        }
    }
    //=====================================================================
    void ARProcess_cl::Finalize()
    {
        m_Thread.Join();
        m_Thread.Finalize();

        for ( int idx = 0; idx < 2; idx++ )
        {
            ma_Photo[ idx ].Finalize();
            ma_Info[ idx ].Finalize();
        }
    }

    ARProcess_cl    s_ARProcess;

    //=====================================================================
    /**
     *  GX
     */
    void initialize_gfx()
    {
        s_HeapForGx = reinterpret_cast< uptr >( s_AppHeap.Allocate( c_gl_memory_size ) );

        // OtBbNXp FCRAM  8MB m
        s_RenderSystem.Initialize( s_HeapForGx, c_gl_memory_size );

        s_PgID = glCreateProgram();
        s_ShID = glCreateShader( GL_VERTEX_SHADER );

        size_t work_size    = nn::fs::GetRomRequiredMemorySize( 1, 1 );
        void * work         = s_AppHeap.Allocate( work_size );

        NN_UTIL_PANIC_IF_FAILED( nn::fs::MountRom( 1, 1, work, work_size ) );

        nn::fs::FileReader file( L"rom:/shader.shbin" );
        size_t fileSize = file.GetSize();
        void * p_Buf = s_AppHeap.Allocate( fileSize );

        s32 read = file.Read( p_Buf, fileSize );
        glShaderBinary( 1, &s_ShID, GL_PLATFORM_BINARY_DMP, p_Buf, read );
        file.Finalize();
        s_AppHeap.Free( p_Buf );

        s_AppHeap.Free( work );

        glAttachShader( s_PgID, s_ShID );
        glAttachShader( s_PgID, GL_DMP_FRAGMENT_SHADER_DMP );

        glBindAttribLocation( s_PgID, 0, "aPosition" );
        glBindAttribLocation( s_PgID, 1, "aColor" );

        glLinkProgram( s_PgID );
        glValidateProgram( s_PgID );
        glUseProgram( s_PgID );

        glClearColor( 0.36f, 0.42f, 0.5f, 1.0f );
        glClearDepthf( 1.f );

        glEnable( GL_DEPTH_TEST );
        glEnable( GL_SCISSOR_TEST );
        glEnable( GL_CULL_FACE );
        glDepthFunc( GL_LESS );

        s_RenderSystem.SetClearColor( NN_GX_DISPLAY0, 0.36f, 0.42f, 0.5f, 1.0f );
        s_RenderSystem.SetClearColor( NN_GX_DISPLAY1, 0.36f, 0.42f, 0.5f, 1.0f );

        glUniform3i( glGetUniformLocation( s_PgID, "dmp_TexEnv[0].srcRgb" ), GL_PRIMARY_COLOR, GL_PRIMARY_COLOR, GL_PRIMARY_COLOR );
        glUniform3i( glGetUniformLocation( s_PgID, "dmp_TexEnv[0].srcAlpha" ), GL_PRIMARY_COLOR, GL_PRIMARY_COLOR, GL_PRIMARY_COLOR );
    }

    //=====================================================================
    /**
     *  GXI
     */
    void finalize_gfx()
    {
        glDetachShader( s_PgID, GL_DMP_FRAGMENT_SHADER_DMP );
        glDetachShader( s_PgID, s_ShID );
        glDeleteShader( s_ShID );
        glDeleteProgram( s_PgID );

        s_RenderSystem.Finalize();
        s_AppHeap.Free( reinterpret_cast<void*>( s_HeapForGx ) );
    }

    //=====================================================================
    /**
     *  
     */
    void initialize()
    {
        nn::Result result;

        nn::fs::Initialize();

        // hid ̏
        result = nn::hid::Initialize();
        NN_UTIL_PANIC_IF_FAILED( result );

        // q[v̍쐬
        s_AppHeap.Initialize( nn::os::GetDeviceMemoryAddress(),
                              nn::os::GetDeviceMemorySize(),
                              nn::os::ALLOCATE_OPTION_LINEAR );

        // GX
        initialize_gfx();

        // J
        s_CameraModule.Initialize();

        //  projectionݒ
        nn::camera::StereoCameraCalibrationData carib;
        nn::camera::GetStereoCameraCalibrationData( &carib );
        s_Projection.Set( IMG_WIDTH, IMG_HEIGHT, NEAR, FAR, carib.angleOfViewRight );

        //  Y2R
        Y2RProcess_cl::ss().Initialize();

        //  AR
        s_ARProcess.Initialize();
    }

    //=====================================================================
    /**
     *  I
     */
    void finalize()
    {
        nn::hid::Finalize();
        s_CameraModule.Finalize();
        s_ARProcess.Finalize();
        Y2RProcess_cl::ss().Finalize();
        
        finalize_gfx();
        s_AppHeap.Finalize();       
    }

}

//=====================================================================
/**
 *  C
 */
extern "C" void
nnMain()
{
    // z[j[sł悤 nn::applet::Enable() ̂݌Ă
    // z[j[Jڂd{^NbNAX[vɂ͔Ή
    nn::applet::Enable(false);
    
    // 
    initialize();

    //  O̍sƕԂ
    s_NarTrans.SetMildRotation( true, 0.9f );
    //  kȊȌꍇɑfǐ
    s_NarTrans.SetAccelerationOnMildRotation( true, 0.25f );

    nn::math::MTX44   projMtx;
    s_Projection.GetMTX44( projMtx );

    nn::math::MTX34   transMtx;
    nn::math::VEC3    transVec( 0.f, 2.f, 0.f );
    nn::math::MTX34Translate( &transMtx, &transVec );

    nn::math::MTX34   scaleMtx;
    nn::math::VEC3    scaleVec( 2.f, 2.f, 2.f );
    nn::math::MTX34Scale( &scaleMtx, &scaleVec );

    nn::fnd::TimeSpan frameTS;
    nn::fnd::TimeSpan AREndTS,ARRateTS,ARLastTS;

    // pbh
    nn::hid::PadReader padReader;
    nn::hid::PadStatus padStatus;

    // o}[J[̏
    int markerNum = 0;
    f32 scores[ MARKER_LIMIT_NUM ];

    bool loop = true;
    while ( loop )
    {
        nn::fnd::TimeSpan   frameBegin  = nn::os::Tick::GetSystemCurrent().ToTimeSpan();

        //  pbh֘A

        //  pbhǂݍ
        padReader.ReadLatest( &padStatus );

        //  fobO\摜ɉ\邩ݒ肷
        if ( padStatus.trigger & nn::hid::BUTTON_X )
        {
            s_ViewInfo = ! s_ViewInfo;

            mw::nar::Debug_cl::Switch_e  viewInfo = static_cast< mw::nar::Debug_cl::Switch_e >( mw::nar::Debug_cl::me_ViewClear
                                                                                                | mw::nar::Debug_cl::me_ViewLineCandidates
                                                                                                | mw::nar::Debug_cl::me_ViewCheckMesh
                                                                                                | mw::nar::Debug_cl::me_ViewEdgeSlope ) ;
            if ( s_ViewInfo )
            {
                mw::nar::Debug_cl::GetrSwitch().TurnOn( viewInfo ) ;
            }
            else
            {
                mw::nar::Debug_cl::GetrSwitch().TurnOff( viewInfo ) ;
            }
        }

        // o@̐ؑ
        if ( padStatus.trigger & nn::hid::BUTTON_Y )
        {
            s_ARProcess.SwitchDetectWork();
        }

        if ( s_ViewInfo )
        {
            if      ( padStatus.trigger & nn::hid::BUTTON_UP )
            {
                mw::nar::Debug_cl::GetrSwitch().Toggle( mw::nar::Debug_cl::me_ViewLineCandidates ) ;
            }
            else if ( padStatus.trigger & nn::hid::BUTTON_LEFT )
            {
                mw::nar::Debug_cl::GetrSwitch().Toggle( mw::nar::Debug_cl::me_ViewCheckMesh ) ;
            }
            else if ( padStatus.trigger & nn::hid::BUTTON_RIGHT )
            {
                mw::nar::Debug_cl::GetrSwitch().Toggle( mw::nar::Debug_cl::me_ViewEdgeSlope ) ;
            }
        }

        if ( padStatus.trigger & nn::hid::BUTTON_START )
        {
            break;
        }
        
        //  AR`f[^Ă
        if ( s_ARProcess.IsDisplayable() )
        {
            //  ӁFSetRenderTarget́AŁASetViewportsĂB
            s_RenderSystem.SetRenderTarget( NN_GX_DISPLAY0 );
            s_RenderSystem.Clear();
            s_RenderSystem.SetViewport( 0, 40, 240, 320 );
            glScissor( 0, 40, 240, 320 );

            glEnable( GL_BLEND );
            glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );

            //  J摜
            s_ARProcess.GetrDisplayablePhoto().Draw();

            if ( s_ViewInfo )
            {
                //  fobO
                s_ARProcess.GetrDisplayableInfo().Draw();
            }

            mw::nar::MarkerList_ac * p_List = s_ARProcess.GetpDisplayableMarkerList();

            glUseProgram( s_PgID );

            //  projections
            glUniformMatrix4fv( glGetUniformLocation( s_PgID, "uProjection" ), 1, GL_TRUE, static_cast< f32 * >( projMtx ) );

            //  }[J[`
            markerNum = p_List->GetMarkerNum();
            if ( markerNum > 0 )
            {
                for ( int idx = 0 ; idx < markerNum ; idx++ )
                {
                    mw::nar::Marker_cl * p_Marker = p_List->GetpMarker( idx ) ;

                    if ( p_Marker && p_Marker->IsEstimated() )
                    {
                        //  model views
                        nn::math::MTX34   mtx;
                        p_Marker->GetrTransformation().GetMTX34( mtx );

                        nn::math::MTX34Mult( &mtx, &mtx, &transMtx );
                        nn::math::MTX34Mult( &mtx, &mtx, &scaleMtx );
                        nn::math::MTX44   mv( mtx );
                        glUniformMatrix4fv( glGetUniformLocation( s_PgID, "uModelView" ), 1, GL_TRUE, static_cast< f32 * >( mv ) );

                        Cube_cl::draw();

                        scores[idx] = p_Marker->GetScore();
                    }
                }
            }

            s_RenderSystem.SwapBuffers();

            //  `I
            s_ARProcess.DisplayDone();

            AREndTS     = nn::os::Tick::GetSystemCurrent().ToTimeSpan();
            ARRateTS    = AREndTS - ARLastTS;
            ARLastTS    = AREndTS;
        }

        s_RenderSystem.SetRenderTarget( NN_GX_DISPLAY1 );
        s_RenderSystem.Clear();
        s_RenderSystem.SetColor( 1.0f, 1.0f, 1.0f );
        s_RenderSystem.SetFontSize( 8.0f );

        if( s_ARProcess.IsDetectedWorkStrict() )
            s_RenderSystem.DrawText( 10.0f, 20.0f, "= AR sample(Strict) =" );
        else
            s_RenderSystem.DrawText( 10.0f, 20.0f, "= AR sample(Normal) =" );

        s_RenderSystem.DrawText( 10.0f, 30.0f, "    cam: %6lldus", s_ARProcess.GetrDisplayableTransferTime().GetMicroSeconds() );
        s_RenderSystem.DrawText( 10.0f, 40.0f, "     AR: %6lldus", s_ARProcess.GetrDisplayableARTime().GetMicroSeconds() );
        s_RenderSystem.DrawText( 10.0f, 50.0f, "  frame: %6lldus", frameTS.GetMicroSeconds() );
        s_RenderSystem.DrawText( 10.0f, 60.0f, "AR rate: %6lldus", ARRateTS.GetMicroSeconds() );

        for( int idx=0; idx < markerNum; idx++ )
        {
            s_RenderSystem.DrawText( 20.0f, 80.0f + 10.0f * idx, "index%d: %f", idx, scores[idx] );
        }
        s_RenderSystem.SwapBuffers();

        frameTS = nn::os::Tick::GetSystemCurrent().ToTimeSpan() - frameBegin;

        s_RenderSystem.WaitVsync( NN_GX_DISPLAY_BOTH );
    }

    //  I
    finalize();

}

