/*-----------------------------------------------------------------------*
  Project:  Nintendo Augmented Reality Library.
  File:     sampledemos\stereo\main.cpp

  Copyright (C)2011 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo and/or its licensed developers
  and are protected by national and international copyright laws. They
  may not be disclosed to third parties or copied or duplicated in any
  form, in whole or in part, without the prior written consent of
  Nintendo.
  The content herein is highly confidential and should be handled
  accordingly.
*-----------------------------------------------------------------------*/

/**
 *  @file
 *  CTRARgp̊ȒPȃTvłB
 *  @author Satoru Osako
 */
#include <limits>

#include <nn.h>
#include <nn/fnd.h>

#include <nn/gx.h>
#include <nn/math.h>
#include <nn/fs.h>

#include <nn/hid.h>
#include <nn/camera.h>
#include <nn/y2r.h>

#include "nar.h"

#include "texturedRectangle.h"

#include "cube.h"
#include "hatenaMarkerTemplate.h"

#include "narDemoConfig.h"
#include "narDemoCamera.h"
#include "narDemoPhoto.h"
#include "narDemoY2r.h"

#include "demo.h"

#define NEAR        (1)
#define FAR         (5000)

#define MARKER_NUM  (6)

//  RȃTCYɈꂽ}[J[ǂ
#define ALLOW_FREESIZE_MARKER   (1)

namespace
{
    //  摜p[N
    u32 sa_NarImageWork[ NAR_IMAGE_WORK_SIZE_4TH( IMG_WIDTH, IMG_HEIGHT ) ] ;

    //  }[J[op[N
    mw::nar::MarkerDetectWork_cl         s_DetectWork;

    //  }[J[f[^
    mw::nar::MarkerData_tc< MARKER_NUM > sa_MarkerData[ 2 ];

    const int                   c_gl_memory_size = 0x800000;

    //
    demo::RenderSystemExt       s_RenderSystem;

    nn::fnd::ExpHeap            s_AppHeap;
    uptr                        s_HeapForGx;

    mw::nar::MarkerTrans_cl     s_NarTrans;
    mw::nar::Projection_cl      sa_Projection[ 2 ];


    bool    s_ViewInfo = false;

    enum ShaderType
    {
        e_ShaderTypeObject,   //!< IuWFNg`p
        e_ShaderTypeImage,    //!< J摜`p (eNX`\)
        e_ShaderTypeNum
    };

    /* program id */
    GLuint sa_PgID[ e_ShaderTypeNum ];

    /* shader id */
    GLuint sa_ShID[ e_ShaderTypeNum ];

    /// BẽY(ms)
    const int c_ShootingTimePlayMilliSeconds    = 200;

    /// J̐p̔
    const f32 c_HalfCameraAOVHorizontal = 27.45f;

    /// J̊Ԃ̋(cm)
    const f32 c_CameraDistance          = 3.5f;

    /// XN[̋(cm)
    const f32 c_DefaultScreenDistance   = 35.f;

    //=====================================================================
    /*  
     */
    class Line_cl
    {
    public:
        void    Init( const nn::math::VEC2 & cr_A, const nn::math::VEC2 & cr_B )
        {
            m_A = cr_B.y - cr_A.y;
            m_B = cr_A.x - cr_B.x;
            m_C = - m_B * cr_A.y - m_A * cr_A.x;
        }

        bool    GetIntersection( nn::math::VEC2 & r_PI, const Line_cl &cr_Line )
        {
            register f32 dest = m_A * cr_Line.m_B - cr_Line.m_A * m_B;

            if ( dest == 0 )
            {   // 2͕s
                return false;
            }
            else
            {
                register f32 invDest = static_cast< f32 >( 1.0 ) / dest;
                r_PI.x = ( m_B * cr_Line.m_C - cr_Line.m_B * m_C ) * invDest;
                r_PI.y = ( m_C * cr_Line.m_A - cr_Line.m_C * m_A ) * invDest;
            }
            return true;
        }

    private:
        f32 m_A;
        f32 m_B;
        f32 m_C;

    };
    //=====================================================================
    /*  XeI\pݒ
     */
    class StereoSetting_cl
    {
    public:
        enum
        {
            me_LT,
            me_LB,
            me_RB,
            me_RT,
            me_VtxNum
        };

        enum
        {
            me_CamR,
            me_CamL,
            me_CamNum
        };

        StereoSetting_cl()
        {}

        /** 
         */
        void initialize( const nn::camera::StereoCameraCalibrationData & r_Calib )
        {
            //  EJ摜̒_ʒuxNg
            //  Pł̑Ίp̔̒
            f32 hd  = nn::math::TanDeg( r_Calib.angleOfViewRight * 0.5f );
            f32 t   = hd * 0.2f;    //  0.2 = 1 / sqrt( 4 ^ 2 + 3 ^ 2 )
            f32 hw  = t * 4;
            f32 hh  = t * 3;

            //  1mm̃sNZ
            f32 pixelsPerMM             = r_Calib.imageWidth / ( hw * r_Calib.distanceToChart * 2 );
            f32 distanceCamerasInPixel  = -pixelsPerMM * r_Calib.distanceCameras;

            maa_CameraImgVec[ me_CamR ][ me_LT ].Set( -hw,  hh );
            maa_CameraImgVec[ me_CamR ][ me_LB ].Set( -hw, -hh );
            maa_CameraImgVec[ me_CamR ][ me_RB ].Set(  hw, -hh );
            maa_CameraImgVec[ me_CamR ][ me_RT ].Set(  hw,  hh );

            //  t덷̂Asړ́AɔׂďȂ̂ŖłƉ肷
            //  J̌(Y])
            f32 errorXInMM      = ( r_Calib.translationX - distanceCamerasInPixel ) / pixelsPerMM;
            f32 thetaYInDeg     = nn::math::AtanDeg( -errorXInMM / r_Calib.distanceToChart );

            //  J̌(X])
            f32 errorYInMM      = r_Calib.translationY / pixelsPerMM;
            f32 thetaXInDeg     = nn::math::AtanDeg( errorYInMM / r_Calib.distanceToChart );

            //  X,Y,Z]̂ɂ
            nn::math::MTX34RotXYZDeg( &m_CorrectionMtx, thetaXInDeg, thetaYInDeg, r_Calib.rotationZ );

            //  J摜̒_ʒuxNg
            hd  = nn::math::TanDeg( r_Calib.angleOfViewLeft * 0.5f );
            t   = hd * 0.2f;    //  0.2 = 1 / sqrt( 4 ^ 2 + 3 ^ 2 )
            hw  = t * 4;
            hh  = t * 3;

            nn::math::VEC3  a_LCameraImgVec[ 5 ];

            a_LCameraImgVec[ me_LT ].Set( -hw,  hh, -1.f );
            a_LCameraImgVec[ me_LB ].Set( -hw, -hh, -1.f );
            a_LCameraImgVec[ me_RB ].Set(  hw, -hh, -1.f );
            a_LCameraImgVec[ me_RT ].Set(  hw,  hh, -1.f );
            a_LCameraImgVec[ 4 ].Set( 0.f, 0.f, -1.f ); //  center

            //  g~Op
            m_LeftLB.Set( -hw, -hh );
            m_LeftRT.Set(  hw,  hh );

            //  J̒_ʒuxNg
            //  ϊ
            nn::math::VEC3TransformNormalArray( a_LCameraImgVec, &m_CorrectionMtx, a_LCameraImgVec, 5 );

            f32 invZ;
            for ( int vtx = 0; vtx < 4; vtx++ )
            {
                invZ    = 1.f / nn::math::FAbs( a_LCameraImgVec[ vtx ].z );
                maa_CameraImgVec[ me_CamL ][ vtx ].Set( a_LCameraImgVec[ vtx ].x * invZ, a_LCameraImgVec[ vtx ].y * invZ );
            }

            invZ    = 1.f / nn::math::FAbs( a_LCameraImgVec[ 4 ].z );
            m_LCameraCenterVec.Set( a_LCameraImgVec[ 4 ].x * invZ, a_LCameraImgVec[ 4 ].y * invZ );
        }

        /** Be摜ˈʒuAItZbgAXP[AˉesvZ
         *  @param  screenDist
         */
        void calc( f32 screenDist )
        {
            //  Be摜ˈʒuAItZbgAXP[AˉesvZ
            if ( screenDist <= 0.f )    screenDist  = c_DefaultScreenDistance;

            //  g~Oʂ̌vZ

            //  EJscreenDistꂽʏɓeꂽJ摜l

            //  J摜4_AJ̊_̃xNgŕ\
            nn::math::VEC2  aa_CameraImgVtx[ 2 ][ 4 ];

            //  screenDisẗʒuœ˂ꂽۂ̒_vZ
            //  EJ
            int vtx;
            for ( vtx = 0; vtx < 4; vtx++ )
            {
                //  XP[O
                nn::math::VEC2Scale( &aa_CameraImgVtx[ me_CamR ][ vtx ], &maa_CameraImgVec[ me_CamR ][ vtx ], screenDist );
            }

            //  J
            for ( vtx = 0; vtx < 4; vtx++ )
            {
                //  XP[O
                nn::math::VEC2Scale( &aa_CameraImgVtx[ me_CamL ][ vtx ], &maa_CameraImgVec[ me_CamL ][ vtx ], screenDist );
                //  sړ
                aa_CameraImgVtx[ me_CamL ][ vtx ].x -= c_CameraDistance;
            }

            //  windowvZ
            //  J̒S
            nn::math::VEC2  center;
            nn::math::VEC2Scale( &center, &m_LCameraCenterVec, screenDist );
            //  sړ
            center.x -= c_CameraDistance;
            //  EJ摜̒(0,0)
            nn::math::VEC2Scale( &center, &center, 0.5f );

            //  Ίp
            Line_cl a_Diagonal[ 2 ];

            nn::math::VEC2  corner( center.x + TOP_LCD_WIDTH, center.y + TOP_LCD_HEIGHT );

            a_Diagonal[ 0 ].Init( center, corner );

            corner.Set( center.x + TOP_LCD_WIDTH, center.y - TOP_LCD_HEIGHT );

            a_Diagonal[ 1 ].Init( center, corner );

            //  J摜
            Line_cl aa_CameraImgEdgeLine[ 2 ][ 4 ];
            f32 dist    = std::numeric_limits< f32 >::max();
            nn::math::VEC2  point;
            for ( s32 cam = 0; cam < 2; cam++ )
            {
                //  J摜̃GbW
                aa_CameraImgEdgeLine[ cam ][ 0 ].Init( aa_CameraImgVtx[ cam ][ 0 ], aa_CameraImgVtx[ cam ][ 1 ] );
                aa_CameraImgEdgeLine[ cam ][ 1 ].Init( aa_CameraImgVtx[ cam ][ 1 ], aa_CameraImgVtx[ cam ][ 2 ] );
                aa_CameraImgEdgeLine[ cam ][ 2 ].Init( aa_CameraImgVtx[ cam ][ 2 ], aa_CameraImgVtx[ cam ][ 3 ] );
                aa_CameraImgEdgeLine[ cam ][ 3 ].Init( aa_CameraImgVtx[ cam ][ 3 ], aa_CameraImgVtx[ cam ][ 0 ] );
                for ( s32 diag = 0; diag < 2; diag++ )
                {
                    for ( s32 line = 0; line < 4; line++ )
                    {
                        //  Ίpƌ_Ƃ
                        nn::math::VEC2  pi;
                        a_Diagonal[ diag ].GetIntersection( pi, aa_CameraImgEdgeLine[ cam ][ line ] );
                        nn::math::VEC2Sub( &pi, &pi, &center );
                        f32 lenSq = pi.LenSq();
                        if ( lenSq < dist )
                        {
                            dist    = lenSq;
                            point   = pi;
                        }
                    }
                }
            }

            f32 hw = nn::math::FAbs( point.x );
            f32 hh = nn::math::FAbs( point.y );

            //  EJ̃g~OvZ
            //  EJ͉摜̒SƃvWFNV̒S
            nn::math::VEC2  lb( center.x - hw, center.y - hh );
            nn::math::VEC2  rt( center.x + hw, center.y + hh );


            f32 l   = lb.x / nn::math::FAbs( aa_CameraImgVtx[ me_CamR ][ me_LB ].x );
            f32 r   = rt.x / nn::math::FAbs( aa_CameraImgVtx[ me_CamR ][ me_RT ].x );
            f32 b   = lb.y / nn::math::FAbs( aa_CameraImgVtx[ me_CamR ][ me_LB ].y );
            f32 t   = rt.y / nn::math::FAbs( aa_CameraImgVtx[ me_CamR ][ me_RT ].y );

            sa_Projection[ 0 ].GetTrimmed( ma_Proj[ 0 ], l, r, b, t, NEAR, FAR );

            //  J̃g~OvZ
            f32 invDist = 1.f / screenDist;
            l   = ( lb.x + c_CameraDistance ) * invDist / nn::math::FAbs( m_LeftLB.x );
            r   = ( rt.x + c_CameraDistance ) * invDist / nn::math::FAbs( m_LeftRT.x );
            b   = lb.y * invDist / nn::math::FAbs( m_LeftLB.y );
            t   = rt.y * invDist / nn::math::FAbs( m_LeftRT.y );

            sa_Projection[ 1 ].GetTrimmed( ma_Proj[ 1 ], l, r, b, t, NEAR, FAR );
        }

        nn::math::VEC2      m_LeftLB;
        nn::math::VEC2      m_LeftRT;
        nn::math::VEC2      maa_CameraImgVec[ 2 ][ 4 ];
        nn::math::VEC2      m_LCameraCenterVec;
        nn::math::MTX44     ma_Proj[ 2 ];
        nn::math::MTX34     m_CorrectionMtx;
    };

    //=====================================================================
    /**
     *  JW[
     */
    class CameraModule_cl
    {
    public:
        void    Initialize();
        void    Finalize();

        /** 2̃J̉摜̂҂
         */
        bool    WaitStereo();

        CameraProcess_cl &  GetrCameraProcess( u32 idx ) { return ma_Processes[ idx ]; }
    private:
        enum
        {
            me_EventCamera1,
            me_EventCamera2,
            me_EventEnd,
            me_EventNum
        };

        CameraProcess_cl        ma_Processes[ 2 ];
        nn::os::WaitObject *    mpa_WaitObjects[ me_EventNum ];
        nn::os::Event           m_EndEvent;
    };
    //=====================================================================
    void CameraModule_cl::Initialize()
    {
        // CAMERA 
        nn::camera::Initialize();

        // obt@ɂ߂郉Cw
        u32 transferLine = nn::camera::GetMaxLines( IMG_WIDTH, IMG_HEIGHT );

        if ( IMG_WIDTH < 16 ) // GetMaxLines͍ő僉C擾֐Ȃ̂ŁAX̃Cȉ̏ꍇ蓾B
        {
            transferLine = IMG_HEIGHT;
        }

        //  Nx
        nn::camera::SetSharpness( nn::camera::SELECT_OUT1_OUT2, 3 );

        //  Io
        nn::camera::SetExposure( nn::camera::SELECT_OUT1_OUT2, 3 );

        //  t[[g̕ύX
        nn::camera::SetFrameRate( nn::camera::SELECT_OUT1_OUT2, nn::camera::FRAME_RATE_30 );

        //  摜TCY̕ύX
        nn::camera::SetSize( nn::camera::SELECT_OUT1_OUT2, CAMERA_SIZE, nn::camera::CONTEXT_A );

        //  mCYtB^(XeIJ̎̓It)
        nn::camera::SetNoiseFilter( nn::camera::SELECT_OUT1_OUT2, false );

        //  XeIJ̎Io͍EJœƗɂ͂炫܂(܂AAĂ܂)B
        //  Io̒x͉摜̖邳ɊÂČvZ܂AẼJɎʂ镗iɂ͈Ⴂ邽߁A
        //  摜Ŝ̖邳玩Iô͂炫߂悤ɂƂAẺ摜̖邳傫قȂ\܂B
        //  ŁAIô͂炫߂邽߂̗̈AẼJɎʂ̈A
        //  ܂AJł͉Ë̗AEJł͍̗̈ɐݒ肷邱ƂŁA邳̈Ⴂ}܂B
        nn::camera::SetAutoExposureWindow( nn::camera::SELECT_OUT1, 0, 0, 480, 480 );
        nn::camera::SetAutoExposureWindow( nn::camera::SELECT_OUT2, 160, 0, 480, 480 );

        //  IoƓlɁAI[gzCgoX̂͂炫߂邽߂̗̈ݒ肵܂B
        //  ɂāAI[gzCgoX̂͂炫̈ႢŔAF̈Ⴂ}܂B
        nn::camera::SetAutoWhiteBalanceWindow( nn::camera::SELECT_OUT1, 0, 0, 480, 480 );
        nn::camera::SetAutoWhiteBalanceWindow( nn::camera::SELECT_OUT2, 160, 0, 480, 480 );

        //  JNԂɂ
        nn::camera::Activate( nn::camera::SELECT_OUT1_OUT2 );

        //  JvZXN
        ma_Processes[ 0 ].Initialize( nn::camera::PORT_CAM1, transferLine, &s_AppHeap );
        ma_Processes[ 1 ].Initialize( nn::camera::PORT_CAM2, transferLine, &s_AppHeap );

        //  JvZX̃Cxg擾
        mpa_WaitObjects[ 0 ] = ma_Processes[ 0 ].GetpCapturedEvent();
        mpa_WaitObjects[ 1 ] = ma_Processes[ 1 ].GetpCapturedEvent();

        //  IpCxg
        m_EndEvent.Initialize( false );
        mpa_WaitObjects[ me_EventEnd ] = &m_EndEvent;

        //  vsync
        nn::camera::SynchronizeVsyncTiming( nn::camera::SELECT_OUT1, nn::camera::SELECT_OUT2 );

        //  JvZXX^[g
        ma_Processes[ 0 ].StartThread();
        ma_Processes[ 1 ].StartThread();
    }
    //=====================================================================
    bool CameraModule_cl::WaitStereo()
    {
        do
        {
            //  ǂ炩Beꂽ^C~OŃ`FbN
            s32 id = nn::os::WaitObject::WaitAny( mpa_WaitObjects, me_EventNum );

            //  Iɐi
            if ( id == me_EventEnd )
                return false;

            //  2̃J̎Be^CX^v𓾂B
            //  ႢȂ΁AɐiށB

            //  ŐṼLv`ʂbN
            ma_Processes[ 0 ].LockImage();
            ma_Processes[ 1 ].LockImage();

            int st0 = static_cast< int >( ma_Processes[ 0 ].GetrShootingTimeOfLockedImage().GetMilliSeconds() );
            int st1 = static_cast< int >( ma_Processes[ 1 ].GetrShootingTimeOfLockedImage().GetMilliSeconds() );

            int diff = abs( st0 - st1 );
            if ( diff < c_ShootingTimePlayMilliSeconds )
            {
                return true;
            }
        }
        while ( true );
    }
    //=====================================================================
    void CameraModule_cl::Finalize()
    {
        ma_Processes[ 0 ].Finalize();
        ma_Processes[ 1 ].Finalize();
        m_EndEvent.Signal();

        nn::camera::Finalize();
    }

    CameraModule_cl s_CameraModule;

    //=====================================================================
    /**
     *  ARvZX
     */
    class ARProcess_cl
    {
    public:
        ARProcess_cl()
             : m_Displaying( 0 ), m_Processing( 0 ), m_Displayable( false )
        {}

        void    Initialize();
        void    Finalize();
        bool    IsDisplayable() { return m_Displayable; }
        void    DisplayDone() { m_Displayable = false; }

        Photo_cl &      GetrDisplayablePhoto( u32 cam )          { return maa_Photo[ cam ][ m_Displaying ]; }
        Photo_cl &      GetrDisplayableInfo( u32 cam )           { return maa_Info[ cam ][ m_Displaying ]; }

        mw::nar::MarkerList_ac *    GetpDisplayableMarkerList( u32 cam )    { return mpaa_MarkerLists[ cam ][ m_Displaying ]; }

        nn::fnd::TimeSpan & GetrDisplayableTransferTime( u32 cam )   { return maa_TransferTime[ cam ][ m_Displaying ]; }
        nn::fnd::TimeSpan & GetrDisplayableARTime( u32 cam )         { return maa_ARTime[ cam ][ m_Displaying ]; }
    private:
        s8          m_Displaying;
        s8          m_Processing;
        bool        m_Displayable;
        u8          a_Pad[1];

        Photo_cl                    maa_Photo[ 2 ][ 2 ];
        Photo_cl                    maa_Info[ 2 ][ 2 ];
        mw::nar::MarkerList_ac *    mpaa_MarkerLists[ 2 ][ 2 ];
        nn::os::Thread              m_Thread;
        u32                         pad;
        nn::fnd::TimeSpan           maa_TransferTime[ 2 ][ 2 ];
        nn::fnd::TimeSpan           maa_ARTime[ 2 ][ 2 ];


        static void hook( ARProcess_cl * p )
        {
            p->process();
        }
        void    process();
    };
    //=====================================================================
    void ARProcess_cl::Initialize()
    {
        for ( int idx = 0; idx < 2; idx++ )
        {
            for ( int cam = 0; cam < 2; cam++ )
            {
                maa_Photo[ idx ][ cam ].Initialize( &s_RenderSystem, &s_AppHeap );
                maa_Info [ idx ][ cam ].Initialize( &s_RenderSystem, &s_AppHeap );
            }
        }

        //  CXbhႢ
        m_Thread.StartUsingAutoStack( hook, this, 4096, nn::os::Thread::GetCurrentPriority() + 2 );
    }
    //=====================================================================
    void ARProcess_cl::process()
    {
        mw::nar::MarkerDatabase_cl   narDatabase;
        mw::nar::Image_cl            narImage( IMG_WIDTH, IMG_HEIGHT, sa_NarImageWork ) ;

        narDatabase.Register( g_HatenaMarker ) ;

        while ( true )
        {
            //  J摜̂҂
            if ( !s_CameraModule.WaitStereo() )
                break;

            //  Y2Rϊsčs

            //  Y2RWuo^
            Y2RJob_cl   a_Job[ 2 ];
            int cam = 0;
            for ( ; cam < 2; cam++ )
            {
                CameraProcess_cl & r_Process = s_CameraModule.GetrCameraProcess( cam );
                a_Job[ cam ].Set( r_Process.GetpLockedYuv(), maa_Photo[ cam ][ m_Processing ].GetpImage() );
                Y2RProcess_cl::ss().Push( a_Job[ cam ] );
            }

            //  Y2RϊJn
            Y2RProcess_cl::ss().Start();

            //  AR}[J[o
            for ( cam = 0; cam < 2; cam++ )
            {
                nn::fnd::TimeSpan   ARBegin = nn::os::Tick::GetSystemCurrent().ToTimeSpan();

                CameraProcess_cl & r_Process = s_CameraModule.GetrCameraProcess( cam );
                narImage.SetImage( r_Process.GetpLockedYuv() );
                narImage.SetDebugImage( maa_Info[ cam ][ m_Processing ].GetpImage() );

                //  }[J[o
                int markerNum = mw::nar::MarkerDetector_cl::Detect( narImage, narDatabase, sa_MarkerData[ cam ], s_DetectWork );

                //  AR}[J[W菈
                for ( int idx = 0 ; idx < markerNum ; idx++ )
                {
                    mw::nar::Marker_cl * p_Marker = sa_MarkerData[ cam ].GetrDetectingMarkerList().GetpMarker( idx ) ;

                    //  k}
                    p_Marker->SuppressShake( 4.f, 16.f );

                    if ( ! p_Marker->IsEstimated() )
                    {
                        // }[J[Wn̐
                        s_NarTrans.Estimate( *p_Marker, sa_Projection[ cam ] );
                    }
                }

                maa_ARTime[ cam ][ m_Processing ]   = nn::os::Tick::GetSystemCurrent().ToTimeSpan() - ARBegin;

                //  ]ԕۑ
                maa_TransferTime[ cam ][ m_Processing ] = r_Process.GetrTransferTimeOfLockedImage();
                //  oʂւ̃|C^ێ
                mpaa_MarkerLists[ cam ][ m_Processing ] = &( sa_MarkerData[ cam ].GetrDetectingMarkerList() );

            }

            //  Y2Rϊ҂
            Y2RProcess_cl::ss().WaitJobFinish();

            m_Displaying    = m_Processing;
            //  ̃vZXCfbNX
            m_Processing ^= 0x1;

            m_Displayable   = true;
        }
    }
    //=====================================================================
    void ARProcess_cl::Finalize()
    {
        m_Thread.Join();
        m_Thread.Finalize();

        for ( int idx = 0; idx < 2; idx++ )
        {
            for ( int cam = 0; cam < 2; cam++ )
            {
                maa_Photo[ idx ][ cam ].Finalize();
                maa_Info [ idx ][ cam ].Finalize();
            }
        }
    }
    
    ARProcess_cl    s_ARProcess;
    
    /**
     *  Ẽ}[J[񂩂K؂ȎpƉ摜̒sNX
     *
     *  Б̃}[J[̎p𕽍sړēo邱ƂŁAẼf
     *  ̕\px傫قȂɑΉĂ
     */
    class StereoMarkerInfo_cl
    {
    public:
        StereoMarkerInfo_cl()
        : m_ID( -1 ), m_ScreenDist( 0.f ), m_EyeWidth( 0.f ),
          m_PrevScreenDist( c_DefaultScreenDistance ), m_PrevEyeWidth( c_CameraDistance )
        {}
        
        /**
         * }[J[̍E̎pvZāA}[J[̈ʒuŉ摜
         * dȂ悤ɁA摜̈ʒuƑ傫𒲐
         */
        void calc( StereoSetting_cl & r_SS )
        {
            int mainEye     = -1;
            int mainIdx     = -1;
            int subEye      = -1;
            int subIdx      = -1;
            
            // Zol̃Zbg
            m_ID   = -1;
            m_EyeWidth   = m_PrevEyeWidth;
            m_ScreenDist = m_PrevScreenDist;
            
            // ̃Tvł́AJ̉摜D悵āAłXRA̍}[J[
            // P̗pĂBJ̉摜Ƀ}[J[ꍇ͉EڂTB
            // iČopJŒ肵ĂȂj
            f32 maxScore = 0.f;
            for ( int eye = 1; eye >=0; eye-- )
            {
                mainEye = eye; subEye = eye ^ 0x1;
                mw::nar::MarkerList_ac * p_List = s_ARProcess.GetpDisplayableMarkerList( eye );
                for ( int marker = 0; marker < p_List->GetMarkerNum(); marker++ )
                {
                    mw::nar::Marker_cl * p_Marker = p_List->GetpMarker( marker );
                    if( p_Marker->IsEstimated() && maxScore < p_Marker->GetScore() )
                    {
                        maxScore = p_Marker->GetScore();
                        mainIdx  = marker;
                    }
                }
                if( mainIdx >= 0 )
                {
                    break; // Ōꍇ̓uCN
                }
            }

            if( mainIdx >= 0 )
            {
                // ɁAΉ}[J[TuJŔF}[J[T
                mw::nar::MarkerList_ac * p_ListMain = s_ARProcess.GetpDisplayableMarkerList( mainEye );
                mw::nar::MarkerList_ac * p_ListSub  = s_ARProcess.GetpDisplayableMarkerList( subEye  );
                
                mw::nar::Marker_cl *         p_MarkerM = p_ListMain->GetpMarker( mainIdx );
                mw::nar::Transformation_cl & r_TM      = p_MarkerM->GetrTransformation();
                
                f32 eyeWidth = 0.f;
                f32 dist2    = c_DefaultScreenDistance * 2.f;
                f32 minDiff  = std::numeric_limits< f32 >::max();
                
                for ( int markerS = 0; markerS < p_ListSub->GetMarkerNum(); markerS++ )
                {
                    mw::nar::Marker_cl * p_MarkerS    = p_ListSub->GetpMarker( markerS );
                    mw::nar::Transformation_cl & r_TS  = p_MarkerS->GetrTransformation();
                    
                    if ( p_MarkerM->GetID() == p_MarkerS->GetID() )
                    {
                        nn::math::VEC3 vecTL;
                        nn::math::VEC3 vecTR;
                        
                        if( mainEye == 0 )
                        {
                            r_TM.GetrTrans().CopyTo( vecTR );
                            r_TS.GetrTrans().CopyTo( vecTL );
                        }
                        else
                        {
                            r_TM.GetrTrans().CopyTo( vecTL );
                            r_TS.GetrTrans().CopyTo( vecTR );
                        }
                        
                        //	Lu[VΉ
                        nn::math::VEC3TransformNormal( &vecTL, &r_SS.m_CorrectionMtx, &vecTL );
                        
                        // Q̌oHŃJ̈ʒurčW̍ŏɂȂyA𓯂}[J[Ƃ݂ȂB
                        // }[J[TCYwƈقȂꍇ̓Y(d)͑ΓIɑ傫Ȃ邽߁AALLOW_FREESIZE_MARKER
                        // gpقmɌvZł邪A̔r͂ł̂ŏȗĂ
                        f32 d = mw::nar::Pow2( vecTR.x - vecTL.x + c_CameraDistance )
                              + mw::nar::Pow2( vecTR.y - vecTL.y ) + mw::nar::Pow2( vecTR.z - vecTL.z );
                        if ( minDiff > d )
                        {
                            subIdx      = markerS;
                            eyeWidth    = vecTL.x - vecTR.x; // X̍
                            dist2       = vecTR.z + vecTL.z; // ̕ςƂAQȗĂ
                            minDiff     = d;                 // Y̍ŏl
                        }
                    }
                }
                
                // ̃Tvł͂ĂȂAČoЕ̃JɌŒ肵A
                // ̊ԊuŎ擾ڕg΁AC̃J摜ōE̎p
                // 擾ł邽ߏłB
                // Q : AR Cu vO~O}jA 3.8.2
                if ( subIdx>=0 && eyeWidth > 0.f )
                {
                    m_PrevEyeWidth = m_EyeWidth = eyeWidth;
                    
#if ALLOW_FREESIZE_MARKER
                    //  }[J[̑傫wƈقȂꍇl
                    f32 scale = c_CameraDistance / m_EyeWidth;
#else
                    f32 scale = 1.f;
#endif
                    //  Be摜ˈʒu𒲐ăZbg
                    m_PrevScreenDist = m_ScreenDist = dist2 * scale * -0.5f;
                }

                // m肵̂IDZbg
                m_ID = p_MarkerM->GetID();
                
                // }[J[̎pZbg
                p_MarkerM->GetrTransformation().GetMTX34( ma_Mtx[ mainEye ] );
                
                if( mainEye ) // CJ̏ꍇ
                {
                    // Lu[VΉ
                    nn::math::MTX34Mult( &ma_Mtx[ mainEye ], &r_SS.m_CorrectionMtx, &ma_Mtx[ mainEye ] );
                    // EZo
                    nn::math::MTX34MultTranslate( &ma_Mtx[ subEye ], nn::math::VEC3(-m_EyeWidth, 0.f, 0.f), ma_Mtx[ mainEye ] );
                }
                else         // CEJ̏ꍇ
                {
                    // Zo
                    nn::math::MTX34MultTranslate( &ma_Mtx[ subEye ], nn::math::VEC3(m_EyeWidth, 0.f, 0.f), ma_Mtx[ mainEye ] );
                    // Lu[VΉ
                    nn::math::MTX34Mult( &ma_Mtx[ subEye ], &r_SS.m_CorrectionMtx, &ma_Mtx[ subEye ] );
                    // Lu[VɉEēxXV
                    nn::math::MTX34MultTranslate( &ma_Mtx[ mainEye ], nn::math::VEC3(-m_EyeWidth, 0.f, 0.f), ma_Mtx[ subEye ] );
                }
                
            }  // mainIdx >= 0
            
            // J摜̒
            r_SS.calc( m_ScreenDist );
            m_MainEye = mainEye; // mFpɃCJێ
        }
        
        
        s32 GetID()         const   { return m_ID;          }
        s32 GetMainEye()    const   { return m_MainEye;    }
        f32 GetScreenDist() const   { return m_ScreenDist;  }
        f32 GetEyeWidth()   const   { return m_EyeWidth;    }
        nn::math::MTX34 & GetrMtx( s32 eye ){ return ma_Mtx[ eye ];    }
        
    private:
        s32 m_ID;
        s32 m_MainEye;
        f32 m_ScreenDist;
        f32 m_EyeWidth;
        f32 m_PrevScreenDist;
        f32 m_PrevEyeWidth;
        nn::math::MTX34  ma_Mtx[ 2 ];
    };
    
    StereoMarkerInfo_cl s_StereoMarkerInfo;
    
    //=====================================================================
    /**
     *  GX
     */
    void initialize_gfx()
    {
        s_HeapForGx = reinterpret_cast< uptr >( s_AppHeap.Allocate( c_gl_memory_size ) );

        // OtBbNXp FCRAM  8MB m
        s_RenderSystem.Initialize( s_HeapForGx, c_gl_memory_size );

        size_t work_size    = nn::fs::GetRomRequiredMemorySize( 1, 1 );
        void * work         = s_AppHeap.Allocate( work_size );

        // *** [cal] }[JƔ|̕`ŎgVF[_𕪂
        for( int i = 0; i < e_ShaderTypeNum; i++ )
        {
            sa_PgID[ i ] = glCreateProgram();
            sa_ShID[ i ] = glCreateShader( GL_VERTEX_SHADER );
        }

        NN_UTIL_PANIC_IF_FAILED( nn::fs::MountRom( 1, 1, work, work_size ) );

        nn::fs::FileReader file( L"rom:/shader.shbin" );
        size_t fileSize = file.GetSize();
        void * p_Buf = s_AppHeap.Allocate( fileSize );

        s32 read = file.Read( p_Buf, fileSize );

        glShaderBinary( e_ShaderTypeNum, sa_ShID, GL_PLATFORM_BINARY_DMP, p_Buf, read );
        file.Finalize();
        s_AppHeap.Free( p_Buf );

        s_AppHeap.Free( work );

        // *** [cal] }[JƔ|̕`ŎgVF[_𕪂
        for( int i = 0; i < e_ShaderTypeNum; i++ )
        {
            glAttachShader( sa_PgID[ i ], sa_ShID[ i ] );
            glAttachShader( sa_PgID[ i ], GL_DMP_FRAGMENT_SHADER_DMP );

            switch( i )
            {
            case e_ShaderTypeObject:
                {
                    glBindAttribLocation( sa_PgID[ i ], 0, "aPosition" );
                    glBindAttribLocation( sa_PgID[ i ], 1, "aColor" );
                }
                break;

            case e_ShaderTypeImage:
                {
                    glBindAttribLocation( sa_PgID[ i ], 0, "aPosition" );
                    glBindAttribLocation( sa_PgID[ i ], 1, "aTexCoord" );
                }
                break;

            } //switch

            glLinkProgram( sa_PgID[ i ] );
            glValidateProgram( sa_PgID[ i ] );
            glUseProgram( sa_PgID[ i ] );

            glUniform3i( glGetUniformLocation( sa_PgID[ i ], "dmp_TexEnv[0].srcRgb" ), GL_PRIMARY_COLOR, GL_PRIMARY_COLOR, GL_PRIMARY_COLOR );
            glUniform3i( glGetUniformLocation( sa_PgID[ i ], "dmp_TexEnv[0].srcAlpha" ), GL_PRIMARY_COLOR, GL_PRIMARY_COLOR, GL_PRIMARY_COLOR );
        }

        glClearColor( 0.36f, 0.42f, 0.5f, 1.0f );
        glClearDepthf( 1.f );

        glEnable( GL_DEPTH_TEST );
        glEnable( GL_SCISSOR_TEST );
        glEnable( GL_CULL_FACE );
        glDepthFunc( GL_LESS );

        s_RenderSystem.SetClearColor( NN_GX_DISPLAY0, 0.36f, 0.42f, 0.5f, 1.0f );
        s_RenderSystem.SetClearColor( NN_GX_DISPLAY1, 0.36f, 0.42f, 0.5f, 1.0f );
        s_RenderSystem.SetClearColor( NN_GX_DISPLAY0_EXT, 0.36f, 0.42f, 0.5f, 1.0f );

        s_RenderSystem.SetLcdMode( NN_GX_DISPLAYMODE_STEREO );
    }

    //=====================================================================
    /**
     *  GXI
     */
    void finalize_gfx()
    {
        for( int i = 0; i < e_ShaderTypeNum; i++ )
        {
            glDetachShader( sa_PgID[ i ], GL_DMP_FRAGMENT_SHADER_DMP );
            glDetachShader( sa_PgID[ i ], sa_ShID[ i ] );
            glDeleteShader( sa_ShID[ i ] );
            glDeleteProgram( sa_PgID[ i ] );
        }

        s_RenderSystem.Finalize();
        s_AppHeap.Free( reinterpret_cast<void*>( s_HeapForGx ) );
    }

    //=====================================================================
    /**
     *  
     */
    void initialize()
    {
        nn::Result result;

        nn::fs::Initialize();

        // hid ̏
        result = nn::hid::Initialize();
        NN_UTIL_PANIC_IF_FAILED( result );

        // q[v̍쐬
        s_AppHeap.Initialize( nn::os::GetDeviceMemoryAddress(),
                              nn::os::GetDeviceMemorySize(),
                              nn::os::ALLOCATE_OPTION_LINEAR );

        // GX
        initialize_gfx();

        //  J
        s_CameraModule.Initialize();

        //  Y2R
        Y2RProcess_cl::ss().Initialize();

        //  AR
        s_ARProcess.Initialize();
    }

    //=====================================================================
    /**
     *  I
     */
    void finalize()
    {
        nn::hid::Finalize();
        s_CameraModule.Finalize();
        s_ARProcess.Finalize();
        Y2RProcess_cl::ss().Finalize();

        finalize_gfx();
        s_AppHeap.Finalize();
    }
    
    
    //=============================================================================
    /**
     *  eNX`RoCi̐ݒ
     *  eNX`gpꍇARoCi 2 ̃J[\[XɃeNX`w肵A
     *  gpȂꍇAJ[\[XɃvC}J[w肷
     */
    void SetTextureCombiner( const GLuint pid, const bool isUseTexture )
    {
        if( isUseTexture )
        {
            glUniform4f( glGetUniformLocation( pid, "dmp_TexEnv[2].constRgba"), 0.0f, 0.0f, 0.0f, 1.0f );
            glUniform3i( glGetUniformLocation( pid, "dmp_TexEnv[2].srcRgb"),    GL_TEXTURE0, GL_CONSTANT, GL_CONSTANT );
            glUniform3i( glGetUniformLocation( pid, "dmp_TexEnv[2].srcAlpha"),  GL_TEXTURE0, GL_CONSTANT, GL_CONSTANT );
        }
        else
        {
            glUniform3i( glGetUniformLocation( pid, "dmp_TexEnv[2].srcRgb"),   GL_PRIMARY_COLOR, GL_PREVIOUS, GL_PREVIOUS );
            glUniform3i( glGetUniformLocation( pid, "dmp_TexEnv[2].srcAlpha"), GL_PRIMARY_COLOR, GL_PREVIOUS, GL_PREVIOUS );
        }
        glUniform3i( glGetUniformLocation( pid, "dmp_TexEnv[2].operandRgb"),   GL_SRC_COLOR, GL_SRC_COLOR, GL_SRC_COLOR );
        glUniform3i( glGetUniformLocation( pid, "dmp_TexEnv[2].operandAlpha"), GL_SRC_ALPHA , GL_SRC_ALPHA, GL_SRC_ALPHA );
        glUniform1i( glGetUniformLocation( pid, "dmp_TexEnv[2].combineRgb"),   GL_REPLACE );
        glUniform1i( glGetUniformLocation( pid, "dmp_TexEnv[2].combineAlpha"), GL_REPLACE );
        glUniform1f( glGetUniformLocation( pid, "dmp_TexEnv[2].scaleRgb"),     1.0f );
        glUniform1f( glGetUniformLocation( pid, "dmp_TexEnv[2].scaleAlpha"),   1.0f );
    }
}

//=====================================================================
/**
 *  C
 */
extern "C" void
nnMain()
{
    // z[j[sł悤 nn::applet::Enable() ̂݌Ă
    // z[j[Jڂd{^NbNAX[vɂ͔Ή
    nn::applet::Enable(false);
    
    // 
    initialize();

    //  O̍sƕԂ
    s_NarTrans.SetMildRotation( true, 0.9f );
    //  kȊȌꍇɑfǐ
    s_NarTrans.SetAccelerationOnMildRotation( true, 0.25f );

    // XeI摜̃Y₷悤ɕ\f
    // ʂ̃Tv菬Ă
    nn::math::MTX34   transMtx;
    nn::math::VEC3    transVec( 0.f, 1.f, 0.f );
    nn::math::MTX34Translate( &transMtx, &transVec );

    nn::math::MTX34   scaleMtx;
    nn::math::VEC3    scaleVec( 1.f, 1.f, 1.f );
    nn::math::MTX34Scale( &scaleMtx, &scaleVec );

    nn::fnd::TimeSpan   frameTS;
    nn::fnd::TimeSpan   AREndTS, ARRateTS, ARLastTS;

    //  pbh
    nn::hid::PadReader padReader;
    nn::hid::PadStatus padStatus;

    //  XeIJ
    StereoSetting_cl ss;

    nn::camera::StereoCameraCalibrationData  cal;
    nn::camera::GetStereoCameraCalibrationData( &cal );
    sa_Projection[ 0 ].Set( IMG_WIDTH, IMG_HEIGHT, NEAR, FAR, cal.angleOfViewRight );
    sa_Projection[ 1 ].Set( IMG_WIDTH, IMG_HEIGHT, NEAR, FAR, cal.angleOfViewLeft );
    ss.initialize( cal );
    ss.calc( c_DefaultScreenDistance );

    bool loop = true;
    while ( loop )
    {
        nn::fnd::TimeSpan   frameBegin  = nn::os::Tick::GetSystemCurrent().ToTimeSpan();

        //  pbh֘A
        //  pbhǂݍ
        padReader.ReadLatest( &padStatus );

        if ( padStatus.trigger & nn::hid::BUTTON_X )
        {
            s_ViewInfo = ! s_ViewInfo;

            mw::nar::Debug_cl::Switch_e viewInfo = static_cast< mw::nar::Debug_cl::Switch_e >( mw::nar::Debug_cl::me_ViewClear
                                                                                               | mw::nar::Debug_cl::me_ViewLineCandidates
                                                                                               | mw::nar::Debug_cl::me_ViewCheckMesh
                                                                                               | mw::nar::Debug_cl::me_ViewEdgeSlope ) ;
            if ( s_ViewInfo )
            {
                mw::nar::Debug_cl::GetrSwitch().TurnOn( viewInfo ) ;
            }
            else
            {
                mw::nar::Debug_cl::GetrSwitch().TurnOff( viewInfo ) ;
            }
        }

        if ( s_ViewInfo )
        {
            if      ( padStatus.trigger & nn::hid::BUTTON_UP )
            {
                mw::nar::Debug_cl::GetrSwitch().Toggle( mw::nar::Debug_cl::me_ViewLineCandidates ) ;
            }
            else if ( padStatus.trigger & nn::hid::BUTTON_LEFT )
            {
                mw::nar::Debug_cl::GetrSwitch().Toggle( mw::nar::Debug_cl::me_ViewCheckMesh ) ;
            }
            else if ( padStatus.trigger & nn::hid::BUTTON_RIGHT )
            {
                mw::nar::Debug_cl::GetrSwitch().Toggle( mw::nar::Debug_cl::me_ViewEdgeSlope ) ;
            }
        }

        if ( padStatus.trigger & nn::hid::BUTTON_START )
        {
            break;
        }

        //  AR`f[^Ă
        if ( s_ARProcess.IsDisplayable() )
        {
            // Ẽ}[J[摜}[J[̎pZoƉ摜̒
            s_StereoMarkerInfo.calc( ss );
            
            GLenum a_RenderTarget[] = { NN_GX_DISPLAY0_EXT, NN_GX_DISPLAY0 };

            for ( int eye = 0; eye < 2; eye++ )
            {
                //  ӁFSetRenderTarget́AŁASetViewportsĂB
                s_RenderSystem.SetRenderTarget( a_RenderTarget[ eye ] );
                s_RenderSystem.Clear();
                glViewport( 0, 0, 240, 400 );
                glScissor( 0, 0, 240, 400 );

                glEnable( GL_BLEND );
                glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );

                GLuint pgID = 0;

                // *** [cal] 摜̕`
                // RenderSystem g킸ɒڕ`悷

                pgID = sa_PgID[ e_ShaderTypeImage ];

                glUseProgram( pgID );

                //  projections
                glUniformMatrix4fv( glGetUniformLocation( pgID, "uProjection" ), 1, GL_TRUE, static_cast< f32 * >( ss.ma_Proj[ eye ] ) );


                // 摜̃r[Wnɂ`ʒu (`XN[ɒu)
                const float zOffset = -1.0f / nn::math::TanDeg( c_HalfCameraAOVHorizontal );

                // 摜 Y ItZbg
                // 512*384 ̉摜 512*512 ̐^񒆂ɒuĂ̂
                // (512 - 384) / 2 ̕ɂ炳ȂƐ^񒆂\łȂ
                // 摜ł̔䗦ł΁A(512-384)/(512)/2 = 0.125 点΂悢A
                // |̑傫 2 Ȃ̂ 2 { 0.25 ƂȂ
                const float yOffset = -0.25f;

                // model view s
                nn::math::MTX34     imageTransMtx;
                nn::math::VEC3      imageTransVec( 0.f, yOffset, zOffset );
                nn::math::MTX34Translate( &imageTransMtx, &imageTransVec );

                //  EJ͓Ȃ
                if ( eye )
                {
                    nn::math::MTX34Mult( &imageTransMtx, &ss.m_CorrectionMtx, &imageTransMtx );
                }
                nn::math::MTX44 imageMv( imageTransMtx );

                glUniformMatrix4fv( glGetUniformLocation( pgID, "uModelView" ), 1, GL_TRUE, static_cast< f32 * >( imageMv ) );

                // eNX`jbg̐ݒ
                SetTextureCombiner( pgID, true );
                glActiveTexture( GL_TEXTURE0 );
                glUniform1i( glGetUniformLocation( pgID, "dmp_Texture[0].samplerType" ), GL_TEXTURE_2D );

                glDisable( GL_DEPTH_TEST );

                // 摜\|̕`
                u16 * p_Image = s_ARProcess.GetrDisplayablePhoto( eye ).GetpImage();
                TexturedRectangle_cl::draw( eye, p_Image, TEXTURE_WIDTH, TEXTURE_HEIGHT );

                if ( s_ViewInfo )
                {
                    p_Image = s_ARProcess.GetrDisplayableInfo( eye ).GetpImage();
                    TexturedRectangle_cl::draw( eye, p_Image, TEXTURE_WIDTH, TEXTURE_HEIGHT );
                }

                glEnable( GL_DEPTH_TEST );

                //  }[J[cube`
                pgID = sa_PgID[ e_ShaderTypeObject ];

                glUseProgram( pgID );  // vOIuWFNg̐؂ւ

                SetTextureCombiner( pgID, false );

                //  projections
                glUniformMatrix4fv( glGetUniformLocation( pgID, "uProjection" ), 1, GL_TRUE, static_cast< f32 * >( ss.ma_Proj[ eye ] ) );

                //  }[J[`
                if( s_StereoMarkerInfo.GetID() != -1 )
                {
                    //  model views
                    nn::math::MTX34  mtx( s_StereoMarkerInfo.GetrMtx( eye ) ); 
                    nn::math::MTX34Mult( &mtx, &mtx, &transMtx );
                    nn::math::MTX34Mult( &mtx, &mtx, &scaleMtx );
                    nn::math::MTX44   mv( mtx );
                    glUniformMatrix4fv( glGetUniformLocation( pgID, "uModelView" ), 1, GL_TRUE, static_cast< f32 * >( mv ) );
                    Cube_cl::draw();
                }
                
                s_RenderSystem.Transfer();
            }

            s_RenderSystem.SwapBuffers();

            //  `I
            s_ARProcess.DisplayDone();

            AREndTS     = nn::os::Tick::GetSystemCurrent().ToTimeSpan();
            ARRateTS    = AREndTS - ARLastTS;
            ARLastTS    = AREndTS;

        }

        s_RenderSystem.SetRenderTarget( NN_GX_DISPLAY1 );
        s_RenderSystem.Clear();

        s_RenderSystem.SetColor( 1.0f, 1.0f, 1.0f );
        s_RenderSystem.SetFontSize( 8.0f );
        s_RenderSystem.DrawText( 10.0f, 20.0f, "= stereo AR sample =" );
        s_RenderSystem.DrawText( 10.0f, 30.0f, "   camR: %6lldus", s_ARProcess.GetrDisplayableTransferTime( 0 ).GetMicroSeconds() );
        s_RenderSystem.DrawText( 10.0f, 40.0f, "   camL: %6lldus", s_ARProcess.GetrDisplayableTransferTime( 1 ).GetMicroSeconds() );
        s_RenderSystem.DrawText( 10.0f, 50.0f, "   AR R: %6lldus", s_ARProcess.GetrDisplayableARTime( 0 ).GetMicroSeconds() );
        s_RenderSystem.DrawText( 10.0f, 60.0f, "   AR L: %6lldus", s_ARProcess.GetrDisplayableARTime( 1 ).GetMicroSeconds() );
        s_RenderSystem.DrawText( 10.0f, 70.0f, "  frame: %6lldus", frameTS.GetMicroSeconds() );
        s_RenderSystem.DrawText( 10.0f, 80.0f, "AR rate: %6lldus", ARRateTS.GetMicroSeconds() );
        s_RenderSystem.DrawText( 10.0f, 90.0f,  "  eye width: %6.2f", s_StereoMarkerInfo.GetEyeWidth() );
        s_RenderSystem.DrawText( 10.0f, 100.0f, "screen dist: %6.2f", s_StereoMarkerInfo.GetScreenDist() );

        if( s_StereoMarkerInfo.GetMainEye() )
        {
            s_RenderSystem.DrawText( 10.0f, 110.0f, "main eye: L" );
        }else{
            s_RenderSystem.DrawText( 10.0f, 110.0f, "main eye: R" );
        }

        s_RenderSystem.Transfer();
        s_RenderSystem.SwapBuffers();

        frameTS = nn::os::Tick::GetSystemCurrent().ToTimeSpan() - frameBegin;

        s_RenderSystem.WaitVsync( NN_GX_DISPLAY_BOTH );
    }

    //  I
    finalize();
}

